import itertools
import warnings
from pathlib import Path
import pandas as pd
import matplotlib.cm as cm
from matplotlib.colors import to_hex
from georunes.tools.data import unique


_markers = ("o", "s", "p", "*", "<", ">", "^", "8", "h", "H", "+",
            "x", "X", "D", "d", "v", "1", "2" "3", "4", "|", "_", ".", 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)


def check_data(dataset, group_name="group", supp_group=None, supp_vtc=None,
               color_column='color', marker_column='marker', zorder_column='zorder', label_column='label',
               ignore_checking_markers=False):
    if supp_group:
        list_group = [group_name, *supp_group]
    else:
        list_group = group_name

    _vtc = (color_column, marker_column, zorder_column, label_column,)
    if supp_vtc:
        vtc = itertools.chain(_vtc, supp_vtc)
    else:
        vtc = _vtc

    if ignore_checking_markers:
        vtc = tuple(x for x in vtc if x != marker_column)

    if group_name not in dataset.keys():
        raise ValueError("Group name '" + str(group_name) + "' not found in dataset")

    groups = dataset.groupby(list_group)
    for name, group in groups:

        for field in vtc:
            if field in group.columns:

                test_value = list(group[field])[0]

                for val in group[field]:
                    if val != test_value:
                        raise ValueError("Different values for group " + str(name) +
                                         " and field " + str(field) + " : " +
                                         str(test_value) + " and " + str(val))


def _open_file(file, sheet=0, group_name='group', sep=",", delimiter=None):
    fext = Path(file).suffix
    if fext in (".xls", ".xlsx") and sheet == 0:
        warnings.warn("No sheet name specified, the first sheet of the Excel file will be used.")

    # Load data
    if fext in (".xls", ".xlsx"):
        data = pd.read_excel(file, sheet_name=sheet)
    elif fext in (".csv", ".txt"):
        data = pd.read_csv(file, sep=sep, delimiter=delimiter)
    if group_name not in data.columns:
        raise ValueError("The specified 'group_name' parameter is not in the excel sheet.")

    return data


def _save_file(data, file, sheet=0, output_suffix="_updated", sep=","):
    fext = Path(file).suffix
    parent_dir = Path(file).parent
    fname = Path(file).stem
    new_fname = str(parent_dir) + '\\' + fname + output_suffix + fext
    if fext in (".xls", ".xlsx"):
        new_sheet = sheet if sheet != 0 else "Sheet1"
        data.to_excel(new_fname, sheet_name=new_sheet, index=False)
    elif fext in (".csv", ".txt"):
        data.to_csv(new_fname, sep=sep)
    print("Edited file saved as '" + new_fname + "'")


def _check_graphic_preset(preset):
    # Check configuration
    attribs = preset.keys()
    if 'group' not in attribs:
        raise ValueError("The parameter 'group' is missing in the config dictionary.")

    categories = preset['group']  # All the categories to update in file
    if len(unique(categories)) < len(categories):
        raise ValueError("A doublon found in definition of group parameters in configuration.")

    card = len(preset['group'])
    for key in attribs:
        if key not in ('group', 'label') and len(preset[key]) != card:
            msg = "The length of the parameter '" + key + \
                  "' is different from the number of group defined in the config."
            raise ValueError(msg)

    return attribs, categories


def _find_graphics_preset(data, category_sequence, categories, cmap='viridis', pop_marker=True, pop_zorder=False,
                          color_column='color', marker_column='marker', zorder_column='zorder', label_column='label',
                          pop_label=False):
    # Colors data to add
    i_cmap = cm.get_cmap(cmap, len(categories))
    rgbcolors = i_cmap(range(len(categories)))
    colors = [to_hex(i) for i in rgbcolors]

    markers = _markers[0:len(categories)]  # Markers data to add

    # Create dictionaries for data attribution
    dict_colors = dict(zip(categories, colors[0:len(categories)]))
    if pop_marker:
        dict_markers = dict(zip(categories, markers))
    if pop_zorder:
        dict_zorders = dict(zip(categories, range(4, 4 + len(categories))))
    if pop_label:
        dict_labels = dict(zip(categories, categories))

    # Create the new contents of the columns to edit
    color_data, marker_data, zorder_data, label_data = [], [], [], []
    for cat in category_sequence:
        color_data.append(dict_colors[cat])
        if pop_marker:
            marker_data.append(dict_markers[cat])
        if pop_zorder:
            zorder_data.append(dict_zorders[cat])
        if pop_label:
            label_data.append(dict_labels[cat])

    # Populate columns data
    data[color_column] = color_data
    if pop_marker:
        data[marker_column] = marker_data
    if pop_zorder:
        data[zorder_column] = zorder_data
    if pop_label:
        data[label_column] = label_data

    return data


def _data_from_sequence(data, preset, category_sequence,
                        group_name='group', color_column='color', marker_column='marker', zorder_column='zorder', label_column='label', ):
    attribs = preset.keys()
    categories = preset[group_name]

    # Data to add
    if not all([x in preset[group_name] for x in unique(category_sequence)]):
        raise ValueError("Some groups existing in data file miss their configuration.")
    if color_column in attribs:
        colors = preset[color_column]
    if marker_column in attribs:
        markers = preset[marker_column]
    if zorder_column in attribs:
        zorders = preset[zorder_column]
    if label_column in attribs:
        labels = preset[label_column]

    # Create dictionaries for data attribution
    if color_column in attribs:
        dict_colors = dict(zip(categories, colors[0:len(categories)]))
    if marker_column in attribs:
        dict_markers = dict(zip(categories, markers))
    if zorder_column in attribs:
        dict_zorders = dict(zip(categories, zorders))
    if label_column in attribs:
        if labels == 'auto' or labels is None:
            dict_labels = dict(zip(categories, categories))
        else:
            dict_labels = dict(zip(categories, labels))

    # Create the new contents of the columns to edit
    color_data, marker_data, zorder_data, label_data = [], [], [], []
    for unit in category_sequence:
        if color_column in attribs:
            color_data.append(dict_colors[unit])
        if marker_column in attribs:
            marker_data.append(dict_markers[unit])
        if zorder_column in attribs:
            zorder_data.append(dict_zorders[unit])
        if label_column in attribs:
            label_data.append(dict_labels[unit])

    # Populate columns data
    if color_column in attribs:
        data[color_column] = color_data
    if marker_column in attribs:
        data[marker_column] = marker_data
    if zorder_column in attribs:
        data[zorder_column] = zorder_data
    if label_column in attribs:
        data[label_column] = label_data

    return data


def file_create_graphic_preset(file, sheet=0, group_name='group', cmap='viridis', pop_marker=True, pop_zorder=False,
                               pop_label=False, output_suffix="_updated", sep=",", delimiter=None):
    # Open file
    data = _open_file(file, sheet=sheet, group_name=group_name, sep=sep, delimiter=delimiter)

    category_sequence = data[group_name].tolist()  # Column of categories
    categories = unique(category_sequence)  # All the categories to update in file

    data = _find_graphics_preset(data, category_sequence, categories, cmap, pop_marker, pop_zorder, pop_label)

    # Save new file
    _save_file(data, file, sheet=sheet, output_suffix=output_suffix)


def data_create_graphic_preset(data, group_name='group', cmap='viridis', color_column='color', pop_marker=True,
                               pop_zorder=False, pop_label=False):
    category_sequence = data[group_name].tolist()  # Column of categories
    categories = unique(category_sequence)  # All the categories to update in file

    data = _find_graphics_preset(data, category_sequence, categories, cmap=cmap, color_column=color_column,
                                 pop_marker=pop_marker, pop_zorder=pop_zorder, pop_label=pop_label)

    return data


def file_set_graphic_preset(file, preset, sheet=0, group_name='group', output_suffix="cfg_update", sep=",",
                            delimiter=None, ):
    # Check preset
    _check_graphic_preset(preset)

    # Open file
    data = _open_file(file, sheet=sheet, group_name=group_name, sep=sep, delimiter=delimiter)

    category_sequence = data[group_name].tolist()  # Column of categories
    data = _data_from_sequence(data, preset, category_sequence, )

    # Save new file
    _save_file(data, file, sheet=sheet, output_suffix=output_suffix, sep=sep)


def data_set_graphic_preset(data, preset, group_name='group', ):
    # Check preset
    _check_graphic_preset(preset)

    # Load data
    if group_name not in data.columns:
        raise ValueError("The specified 'group_name' parameter is not in the excel sheet.")

    category_sequence = data[group_name].tolist()  # Column of categories
    data = _data_from_sequence(data, preset, category_sequence, )

    # Return data
    return data


def check_geochem_res(reservoir, listing=None):  # Listing for eventually required oxide/element
    chklist = ['alias', 'color']
    if listing is not None:
        chklist = chklist.append(*listing)

    keys = reservoir.keys()
    for el in chklist:
        if el not in keys:
            raise ValueError("Key ", el, "not found in reservoir configuration.")

    return True



import math
import numpy as np
from matplotlib.colors import to_rgba

from georunes.plot.helpers import ArrowDrawerTernary
from georunes.plot.ternary.ternbase import DiagramTernaryBase
from georunes.tools.chemistry import molar_ratio, molar_ratio_specified
from georunes.tools.language import format_chemical_formula as _fml
from georunes.tools.language import get_translator
from georunes.tools.plotting import normalize_marker_size

sin60 = math.sqrt(3) / 2
tan30 = 1 / math.sqrt(3)
tan60 = math.sqrt(3)


# Nesbitt, H.W. and Young, G.M., 1984.
# Prediction of some weathering trends of plutonic and volcanic rocks based on thermodynamic and
# kinetic considerations. Geochimica et cosmochimica acta, 48(7).

class DiagramNesbittYoung(DiagramTernaryBase, ArrowDrawerTernary):
    def __init__(self, datasource, title="Nesbitt and Young diagram",
                 markersize=20, ignore_apatite_correction=False, **kwargs):
        DiagramTernaryBase.__init__(self, datasource=datasource, title=title,
                                    markersize=markersize,
                                    top_label=_fml('A (Al2O3)'), left_label=_fml('CN (CaO+Na2O)'),
                                    right_label=_fml('K (K2O)'),
                                    no_ticks_label=True,
                                    **kwargs)

        self.ignore_apatite_correction = ignore_apatite_correction

    def set_decoration(self, ):
        _ = get_translator(self.lang_cfg)
        # Order for points : right var, top var, left var

        _ax = self.tax.get_axes()

        _ax.plot((0, 50 + 35 * sin60 * tan30), (sin60 * 65, sin60 * 65), linewidth=0.8, color=self.decor_line_col,
                 linestyle="--")
        _ax.plot((0, 50 + 15 * sin60 * tan30), (sin60 * 85, sin60 * 85), linewidth=0.8, color=self.decor_line_col,
                 linestyle="--")
        _ax.plot((0, 50 + 50 * sin60 * tan30), (sin60 * 50, sin60 * 50), linewidth=0.8, color=self.decor_line_col,
                 linestyle="--")

        self.ax.arrow(35, 50 * sin60, (30 * sin60) / tan60, 30 * sin60, linewidth=1, color="#222222",
                      head_width=3, head_length=3)

        _ax.annotate(_("Weak\nweathering"), (3, 45,), fontsize="small", color=self.decor_text_col, fontweight='bold')
        _ax.annotate(_("Intermediate\nweathering"), (3, 60,), fontsize="small", color=self.decor_text_col,
                     fontweight='bold')
        _ax.annotate(_("Strong\nweathering"), (3, 75,), fontsize="small", color=self.decor_text_col, fontweight='bold')
        self.tax.annotate(_("Hornblende"), (3, 16,), color=self.decor_text_col, fontsize="small")
        self.tax.annotate(_("Pyroxene"), (3, 6,), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("Plagioclase"), (3, sin60 * 46,), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("Smectite"), (25, sin60 * 80,), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("Kaolinite, gibbsite, chlorite"), (55, 85,), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("Illite"), (64, int(80 * sin60),), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("Muscovite"), (68, int(72 * sin60),), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("Biotite"), (74, int(58 * sin60),), color=self.decor_text_col, fontsize="small")
        _ax.annotate(_("K-feldspar"), (79, int(50 * sin60),), color=self.decor_text_col, fontsize="small")

        minerals = [
            (0, 8, 92),  # Clinopyroxene
            (0, 18, 82),  # Hornblende
            (0, 50, 50),  # Plagioclase
            (0, 80, 20),  # Smectite
            (0, 100, 0),  # Kaolininte, gibbsite, Chlorite
            (20, 80, 0),  # Illite
            (28, 72, 0),  # Muscovite
            (42, 58, 0),  # Biotite
            (50, 50, 0)  # K-feldspar
        ]

        self.tax.scatter(minerals, marker="o", facecolors="#333333", zorder=4, s=20)
        self.ax.arrow(0, 0, 0, 50 * tan60, head_width=3, head_length=3, fc='k', ec='k')

        # Draw vertical axis ticks
        cia_val = np.arange(0, 101, 10)
        for tick_value in cia_val:
            y_tick = tick_value * 50 * tan60 / 100
            self.ax.plot((0, 0.5), (y_tick, y_tick), linewidth=1, color=self.decor_line_col, )
            _ax.annotate(tick_value, (-3, y_tick - 2,), color=self.decor_text_col, fontsize="small", va="bottom",
                         ha="right")

        _ax.annotate("CIA", (2, 22,), color=self.decor_text_col, rotation=90, )

    def plot(self):
        DiagramTernaryBase.plot_config(self, no_gridline=True, no_ticks=self.no_ticks,
                                       no_ticks_label=self.no_ticks_label)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                top_var = molar_ratio(group['Al2O3'])
                if self.ignore_apatite_correction:
                    left_var = molar_ratio_specified(group['CaO'], "CaO") + molar_ratio(group['Na2O'])
                else:
                    left_var = molar_ratio_specified(group['CaO'] - 55.07 * group["P2O5"] / 41.85, "CaO") + molar_ratio(
                        group['Na2O'])

                right_var = molar_ratio(group['K2O'])

                var_sum = top_var + left_var + right_var
                norm_z = 100 * left_var / var_sum
                norm_x = 100 * right_var / var_sum
                norm_y = 100 * top_var / var_sum

                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                # Convert points to tuples
                points = [(norm_x.get(i), norm_y.get(i), norm_z.get(i)) for i, sample in top_var.items()]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.tax.scatter(points, edgecolors=edge_color,
                                 marker=marker, label=label, facecolors=sample_color, s=size,
                                 zorder=zorder)

                if self.annotation:
                    for i, sample in group[self.top_var].items():
                        self.tax.annotate(group[self.annotation].get(i), (norm_x.get(i), norm_y.get(i)),
                                          fontsize='xx-small')

        self.plot_arrows()
        self.fig.subplots_adjust(right=0.92)
        self.plot_legend()



import tkinter as tk
import numpy as np
import matplotlib.pyplot as plt
from tkinter import ttk
from matplotlib.collections import LineCollection
from pandas import Series
from georunes.petromod.finders.cumulate_finder import CumulateFinder
from georunes.tools.gui.tkwidets.entry_with_callback import EntryWithCallback
from georunes.tools.gui.tkwidets.editable_treeview import EditableFloatTreeview
from georunes.petromod.viewer.tk.base_viewer import BaseViewerTk, total_label_color
from georunes.tools.colors import merge_colormaps
from georunes.tools.data import float_or_zero


class MassBalanceViewerTk(BaseViewerTk):
    def __init__(self, petro_model, parent_conc, child_conc=None,
                 suppl_wm_title=None, alpha_step=0.1, *args, **kwargs):
        self.minerals = petro_model.get_minerals_list()
        self.oxels = petro_model.get_elements_list()
        self.default_parent_conc = parent_conc.copy()
        for oxel in self.oxels:
            if oxel not in self.default_parent_conc.keys():
                raise Exception(
                    "Non Null value is required for all elements in parental composition. Check the element " + oxel + ".")
        self.parent = self.default_parent_conc.copy()
        self.parent_elements = self.default_parent_conc.index.tolist()
        self.child = child_conc
        self.set_supplement_status()
        self.oxelx = self.oxels[0]
        self.oxely = self.oxels[1]
        self.alpha_step = alpha_step
        self.show_infos = True
        BaseViewerTk.__init__(self, petro_model, suppl_wm_title=suppl_wm_title, *args, **kwargs)
        self.init_cumulate_modal()
        self.cumulate_finder_method = 'BVLS'
        self.cumulate_finder_norm = 'euclidian'
        self.fill_frame()
        self.adjust_frame_position()
        self.found_cumulate_row = None

    def set_supplement_status(self):
        self.supplement_status = []

    # init cumulate modal composition
    def init_cumulate_modal(self):
        value_per_mineral = 100 / len(self.minerals)
        self.default_cumulate = {mineral: value_per_mineral for mineral in self.minerals}
        self.cumulate_modal = self.default_cumulate.copy()
        total = 0
        for mineral, val in self.default_cumulate.items():
            total += float_or_zero(val)
        self.solid_total = total

        self.missing_minerals = []
        for mineral in self.cumulate_modal.keys():
            if mineral not in self.minerals:
                self.missing_minerals.append(mineral)
        self.missing_oxels = []
        for element in self.parent_elements:
            if element not in self.model.oxel_list:
                self.missing_oxels.append(element)

    def update_parent_conc(self):
        new_parent_conc = self.parent.copy()
        for label, entry in self.parental_conc_widget:
            new_parent_conc[label.cget("text")] = float(entry.get())
        self.parent = new_parent_conc

    def update_cumulate_modal(self):
        new_mineral_props = {}
        total = 0
        for label, entry in self.cumulate_modal_widgets:
            new_mineral_props[label.cget("text")] = float_or_zero(entry.get())
            total += float_or_zero(entry.get())
        total_new_label, total_color = total_label_color(total)
        self.label_total.config(text=total_new_label, fg=total_color)
        self.cumulate_modal = new_mineral_props
        self.solid_total = total

    def update_oxelx(self, _event):  # update_selected_phase
        self.oxelx = self.oxelx_selector.get()
        self.oxely = self.oxely_selector.get()
        self.refresh_plotting()

    def update_oxely(self, _event):  # update_selected_phase
        self.oxely = self.oxely_selector.get()
        self.refresh_plotting()

    def refresh_model(self):
        self.update_parent_conc()
        self.update_cumulate_modal()
        self.refresh_cumulate_concentrations()

    def refresh_model_and_plot(self):
        self.refresh_model()
        self.refresh_plotting()

    def update_minerals_data(self, event):
        new_data_minerals = self.minerals_tree.data_tab
        self.model.set_minerals_data(new_data_minerals)
        self.refresh_cumulate_concentrations()

    def _reset_cumulate_modal(self):
        for index, (label, entry) in enumerate(self.cumulate_modal_widgets):
            entry.delete(0, tk.END)  # Clear the current content
            if label.cget("text") in self.default_cumulate.keys():
                entry.insert(0, self.default_cumulate[label.cget("text")])
        self.refresh_plotting()

    def event_reset_cumulate_modal(self):
        self._reset_cumulate_modal()

    def update_cumulate_finder_method(self, _event):
        self.cumulate_finder_method = self.method_selector.get()

    def update_cumulate_finder_norm(self, _event):
        self.cumulate_finder_norm = self.norm_selector.get()

    def _find_cumulate_frame(self):
        self.finder_level = tk.Toplevel(self.root)
        self.finder_level.title("Search for a cumulate")
        content = tk.Frame(self.finder_level)
        content.pack(side="top")
        content_left = tk.Frame(content)
        content_left.pack(side="left")
        content_right = tk.Frame(content)
        content_right.pack(side="right")
        bottom = tk.Frame(self.finder_level)
        bottom.pack(side="top")

        # Cumulate in content_left
        tk.Label(content_left, text="Cumulate modal composition :").grid(row=0, column=0, columnspan=2, sticky="ew")
        self.cumulate_finder_widgets = list()
        for index, (mineral, value) in enumerate(self.cumulate_modal.items()):
            label = tk.Label(content_left, text=mineral)
            entry = tk.Entry(content_left, width=10)
            entry.insert(0, str(value))
            self.cumulate_finder_widgets.append((label, entry))
        for index, (label, entry) in enumerate(self.cumulate_finder_widgets):  # +1 in the grid
            label.grid(row=index + 1, column=0, sticky=tk.W, pady=2)
            entry.grid(row=index + 1, column=1, sticky=tk.W, pady=2)

        # Child concentration in content_right
        tk.Label(content_right, text="Residual composition to get :").grid(row=0, column=0, columnspan=2, sticky="ew")
        if self.child is not None:
            comp_to_show = self.child
        else:
            comp_to_show = self.parent()

        self.finder_widgets_child = list()
        for index, (oxel, value) in enumerate(comp_to_show.items()):
            label = tk.Label(content_right, text=oxel)
            entry = tk.Entry(content_right, width=10)
            entry.insert(0, value)
            self.finder_widgets_child.append((label, entry))

        for index, (label, entry) in enumerate(self.finder_widgets_child):  # +1 in the grid
            label.grid(row=index + 1, column=0, sticky=tk.W, pady=2)
            entry.grid(row=index + 1, column=1, sticky=tk.W, pady=2)

        tk.Label(bottom, text="Research method").grid(row=0, column=0)
        methods = ['NNLS', 'BVLS', 'RS', 'GD']
        self.method_selector = ttk.Combobox(bottom, width=8, values=methods)
        self.method_selector.set(self.cumulate_finder_method)
        self.method_selector.grid(row=1, column=0)
        self.method_selector.bind("<<ComboboxSelected>>", self.update_cumulate_finder_method)

        tk.Label(bottom, text="Norm").grid(row=0, column=1)
        methods = ['euclidian', 'max', 'min', 'MAE', 'MSE', 'RMSE', 'SMAPE']
        self.norm_selector = ttk.Combobox(bottom, width=8, values=methods)
        self.norm_selector.set(self.cumulate_finder_norm)
        self.norm_selector.grid(row=1, column=1)
        self.norm_selector.bind("<<ComboboxSelected>>", self.update_cumulate_finder_norm)

        compute_cumulate_btn = tk.Button(bottom, text="Compute cumulate", command=self.event_search_cumulate)
        compute_cumulate_btn.grid(row=0, column=2, rowspan=2, sticky="nsew", padx=4, pady=4)
        self.use_cumulate_btn = tk.Button(bottom, text="Use cumulate", fg='red', command=self.event_use_cumulate)
        self.use_cumulate_btn.grid(row=0, column=3, rowspan=2, sticky="nsew", padx=4, pady=4)

        self.finder_level.grab_set()
        self.finder_level.lift()
        self.finder_level.focus_force()

    def event_find_cumulate_frame(self):
        self._find_cumulate_frame()

    def event_search_cumulate(self):
        finder = CumulateFinder(self.model.minerals_data.T.reset_index(),
                                verbose=1, optimizer=self.cumulate_finder_method, norm=self.cumulate_finder_norm)
        self.child_test = self.parent.copy()
        for label, entry in self.finder_widgets_child:
            self.child_test[label.cget("text")] = float(entry.get())

        df_props, df_suppl, fracts_list = finder.find_cumulate_to_remove(self.parent, self.child_test)

        # Select a proportion of the cumulate to extract (alpha)
        select_alpha_level = tk.Toplevel(self.finder_level)
        select_alpha_level.grab_set()
        tk.Label(select_alpha_level, text="Choose a cumulate proportion value :").pack(pady=(0, 10))
        combo = ttk.Combobox(select_alpha_level, values=fracts_list, state="readonly")
        combo.pack()
        combo.current(0)  # Select the first item by default

        def on_ok():
            value = combo.get()
            index = fracts_list.index(float(value)) if float(value) in fracts_list else None

            # Get a composition by the index of the alpha value
            row_selected = df_props.iloc[index][:-1]
            row_selected_dict = row_selected.to_dict()
            self.found_cumulate_row = row_selected_dict

            # Update the displayed infos in cumulate finder toplevel
            for index, (label, entry) in enumerate(self.cumulate_finder_widgets):
                entry.delete(0, tk.END)  # Clear the current content
                if label.cget("text") in row_selected_dict.keys():
                    entry.insert(0, row_selected_dict[label.cget("text")])

            select_alpha_level.destroy()
            self.use_cumulate_btn.config(fg= "green")

        tk.Button(select_alpha_level, text="OK", command=on_ok).pack(pady=(10, 0))
        self.finder_level.wait_window(select_alpha_level)

    def event_use_cumulate(self):
        if self.found_cumulate_row:
            for index, (label, entry) in enumerate(self.cumulate_modal_widgets):
                entry.delete(0, tk.END)  # Clear the current content
                if label.cget("text") in self.found_cumulate_row.keys():
                    entry.insert(0, self.found_cumulate_row[label.cget("text")])
        self.child = self.child_test.copy()

        self.finder_level.destroy()
        self.refresh_plotting()

    def _reset_parent_conc(self):
        for index, (label, entry) in enumerate(self.parental_conc_widget):
            entry.delete(0, tk.END)  # Clear the current content
            entry.insert(0, self.default_parent_conc[label.cget("text")])  # Insert the new content

    def reset_parent_conc(self):
        self._reset_parent_conc()
        self.refresh_model()
        self.refresh_plotting()

    def _reset_minerals_data(self):
        self.model.reset_minerals_data()
        self.refresh_minerals_treeview()

    def event_reset_minerals_data(self):
        self._reset_minerals_data()
        self.refresh_plotting()

    def refresh_minerals_treeview(self):
        for row in self.minerals_tree.get_children():
            self.minerals_tree.delete(row)
        for index, row in self.model.minerals_data.iterrows():
            self.minerals_tree.insert("", tk.END, iid=index, values=[index] + row.tolist())

    def refresh_cumulate_concentrations(self):
        for row in self.cumulate_modal_tree.get_children():
            self.cumulate_modal_tree.delete(row)
        self.cumulate_conc = self.model.modal_props_to_concentration(self.cumulate_modal)
        for oxel, val in self.cumulate_conc.items():
            self.cumulate_modal_tree.insert("", tk.END,tags=('inert',), values=(oxel, val))

    def draw(self, **kwargs):
        self.ax.scatter(self.cumulate_conc[self.oxelx], self.cumulate_conc[self.oxely], marker='o', color='k',
                        label="Cumulate")
        self.ax.scatter(self.parent[self.oxelx], self.parent[self.oxely], marker='s', color='g', label="Parent")
        if self.child is not None:
            self.ax.scatter(self.child[self.oxelx], self.child[self.oxely], marker='D', color='m', label="Child")
        self.draw_model()
        self.ax.set_xlabel(self.oxelx)
        self.ax.set_ylabel(self.oxely)
        if self.show_legend:
            self.ax.legend()
        self.canvas.draw()

    def draw_model(self):
        max_extraction = self.model.compute_alpha_limit_removal(self.parent, Series(self.cumulate_conc))
        fracts = np.arange(0, min(max_extraction, 1.01), self.alpha_step)
        fracts = np.append(fracts, min(max_extraction, 1.01))
        valx = self.model.get_residual_liq_concentration(self.oxelx, self.cumulate_modal, self.parent, fracts)
        valy = self.model.get_residual_liq_concentration(self.oxely, self.cumulate_modal, self.parent, fracts)

        if self.show_infos:
            for i in range(1, len(fracts)):
                self.ax.annotate(str(round(fracts[i], 2)), (valx[i], valy[i]))

        if max_extraction < 1 - self.alpha_step:
            s_fracts = np.arange(fracts[-2] + self.alpha_step, 1.0, self.alpha_step)
            s_valx = self.model.get_residual_liq_concentration(self.oxelx, self.cumulate_modal, self.parent, s_fracts)
            s_valy = self.model.get_residual_liq_concentration(self.oxely, self.cumulate_modal, self.parent, s_fracts)

        else:
            s_fracts = np.arange(fracts[-2] + self.alpha_step, 1.0, self.alpha_step)
            s_valx = self.model.get_residual_liq_concentration(self.oxelx, self.cumulate_modal, self.parent, s_fracts)
            s_valy = self.model.get_residual_liq_concentration(self.oxely, self.cumulate_modal, self.parent, s_fracts)

        all_fracts = [*fracts, *s_fracts]
        all_valx = [*valx, *s_valx]
        all_valy = [*valy, *s_valy]
        cmap = merge_colormaps("winter_r", "autumn", max_extraction)
        self.ax.scatter(all_valx[1:], all_valy[1:], marker='x', cmap=cmap, c=all_fracts[1:], vmin=0, vmax=1)
        all_points = np.array([all_valx, all_valy]).T.reshape(-1, 1, 2)
        all_segments = np.concatenate([all_points[:-1], all_points[1:]], axis=1)
        lc = LineCollection(all_segments, cmap=cmap, norm=plt.Normalize(0, 1))
        lc.set_array(all_fracts[:-1])  # Set which values to color by
        line = self.ax.add_collection(lc)  # Add the gradient line
        self.ax.autoscale()  # Auto-rescale to fit data
        cbar = self.fig.colorbar(line, ax=self.ax)
        cbar.set_label('Extracted fraction')

        if max(all_valx) > 100:
            self.ax.set_xlim(right=99.9)
        if min(all_valx) < -5:
            self.ax.set_xlim(left=-10)
        if max(all_valy) > 100:
            self.ax.set_ylim(top=99.9)
        if min(all_valy) < -5:
            self.ax.set_ylim(bottom=-10)

        if self.show_infos:
            for i in range(0, len(s_fracts)):
                self.ax.annotate(str(round(s_fracts[i], 2)), (s_valx[i], s_valy[i]))

    def fill_frame(self):
        BaseViewerTk.fill_frame(self)

        # Sideview
        self.sideview = tk.Frame(self.content)
        self.sideview.pack(side="left")
        pane_high = tk.Frame(self.sideview, )
        pane_high.pack(side='top')
        pane_low = tk.Frame(self.sideview, )
        pane_low.pack(side='bottom')
        pane_one = tk.Frame(self.sideview, width=140, padx=2, pady=2)
        pane_two = tk.Frame(self.sideview, width=140, padx=2, pady=2)
        pane_three = tk.Frame(self.sideview, width=140, padx=2, pady=2)
        pane_one.pack(side="left", anchor="n")
        pane_two.pack(side="left", anchor="n")
        pane_three.pack(side="left", anchor="n")
        pane_tab = tk.Frame(self.content, width=380, padx=2, pady=2)
        pane_tab.pack_propagate(False)
        pane_tab.pack(side="left", fill=tk.BOTH, anchor="n")

        # In sideview

        # Cumulate modal composition
        comp_title = tk.Label(pane_one, text="Cumulate modal composition", anchor=tk.CENTER)
        comp_title.pack(side="top", )
        ttk.Separator(pane_one, orient='horizontal').pack(fill="x")

        proportions_pane = tk.Frame(pane_one)
        proportions_pane.pack(side="top")
        self.cumulate_modal_widgets = []
        total = 0
        for index, (mineral, value) in enumerate(self.cumulate_modal.items()):
            label = tk.Label(proportions_pane, text=mineral)
            entry = EntryWithCallback(proportions_pane, width=10, callback=self.refresh_model)
            entry.insert(0, str(value))
            entry.enable_callback()  # Enabled after initialization
            self.cumulate_modal_widgets.append((label, entry))
            total += value

        for index, (label, entry) in enumerate(self.cumulate_modal_widgets):
            label.grid(row=index, column=0, sticky=tk.W, pady=2)
            entry.grid(row=index, column=1, sticky=tk.W, pady=2)

        # Find cumulate button
        find_cumulate_btn = tk.Button(pane_one, text="Search cumulate", command=self.event_find_cumulate_frame)
        find_cumulate_btn.pack(side='top')

        # Reset modal composition button
        reset_solid_btn = tk.Button(pane_one, text="Reset solid composition", command=self.event_reset_cumulate_modal)
        reset_solid_btn.pack(side='top')

        # Pane two
        cumulate_modal_title = tk.Label(pane_two, text="Cumulate concentrations", anchor=tk.CENTER)
        cumulate_modal_title.pack(side="top", )

        # Create a Treeview widget to display the DataFrame
        self.cumulate_modal_tree = ttk.Treeview(pane_two, columns=("col1", "col2"),
                                                show="headings")
        self.cumulate_modal_tree.heading("col1", text="Element")
        self.cumulate_modal_tree.heading("col2", text="Concentration")
        self.cumulate_modal_tree.column("col1", width=60, anchor="center")
        self.cumulate_modal_tree.column("col2", width=60, anchor="center")

        self.cumulate_conc = self.model.modal_props_to_concentration(self.cumulate_modal)
        for element, coeff in self.cumulate_conc.items():
            self.cumulate_modal_tree.insert("", "end", tags=('inert',), values=(element, coeff))
        self.cumulate_modal_tree.tag_configure(tagname='inert', background='#f0f0f0')
        self.cumulate_modal_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        # Pane three
        tk.Label(pane_three, text=self.model.init_name, anchor=tk.CENTER).pack(side="top")
        ttk.Separator(pane_three, orient='horizontal').pack(side="top", fill="x")

        concentration_pane = tk.Frame(pane_three)
        concentration_pane.pack(side="top")
        self.parental_conc_widget = []
        for index, (element, value) in enumerate(self.parent.items()):
            label = tk.Label(concentration_pane, text=element)
            entry = tk.Entry(concentration_pane, width=10)
            entry.insert(0, value)
            self.parental_conc_widget.append((label, entry))
        for index, (label, entry) in enumerate(self.parental_conc_widget):
            label.grid(row=index, column=0, sticky=tk.W, pady=2)
            entry.grid(row=index, column=1, sticky=tk.W, pady=2)

        reset_conc_init_btn = tk.Button(pane_three, text="Reset parental concentrations", command=self.reset_parent_conc)
        reset_conc_init_btn.pack()

        # Bottom bar
        ttk.Separator(self.bottombar, orient='horizontal').pack(side="top", fill="x")

        # Oxel selection
        label_oxelx = tk.Label(self.bottombar, text="Oxide/element X", anchor=tk.CENTER)
        label_oxelx.pack(side="left", padx=5)
        self.oxelx_selector = ttk.Combobox(self.bottombar, width=10, values=list(self.oxels))
        self.oxelx_selector.set(self.oxels[0])
        self.oxelx_selector.pack(side="left", padx=5)
        self.oxelx_selector.bind("<<ComboboxSelected>>", self.update_oxelx)
        label_oxely = tk.Label(self.bottombar, text="Oxide/element Y", anchor=tk.CENTER)
        label_oxely.pack(side="left", padx=5)
        self.oxely_selector = ttk.Combobox(self.bottombar, width=10, values=list(self.oxels))
        self.oxely_selector.set(self.oxels[1])
        self.oxely_selector.pack(side="left", padx=5)
        self.oxely_selector.bind("<<ComboboxSelected>>", self.update_oxely)

        ttk.Separator(self.bottombar, orient='vertical').pack(side="left", padx=8, fill="y")
        check_legend = tk.Checkbutton(self.bottombar, text='Show/hide legend', command=self.toggle_legend, )
        check_legend.select()
        check_legend.pack(side="left")

        ttk.Separator(self.bottombar, orient='vertical').pack(side="left", padx=8, fill="y")
        check_infos = tk.Checkbutton(self.bottombar, text='Show/hide information', command=self.toggle_infos)
        check_infos.select()
        check_infos.pack(side="left")

        # Refresh plotting
        ttk.Separator(self.bottombar, orient='vertical').pack(side="left", padx=8, fill="y")
        refresh_plotting_btn = tk.Button(self.bottombar, text="Refresh plotting", command=self.refresh_model_and_plot)
        refresh_plotting_btn.pack(side="left")

        # Status bar
        self.fill_status_bar()

        # Tab view
        # Coefficients Treeview
        label_minerals_tree = tk.Label(pane_tab, text="Minerals compositions", anchor=tk.CENTER)
        label_minerals_tree.pack(side="top")

        # Reset coefficients button in the bottom
        reset_minerals_data_btn = tk.Button(pane_tab, text="Reset mineral compositions",
                                            command=self.event_reset_minerals_data)
        reset_minerals_data_btn.pack(side="bottom", )

        # Treeview
        self.minerals_tree = EditableFloatTreeview(pane_tab, self.model.minerals_data, )
        self.minerals_tree.column('#0', width=0, stretch=tk.NO)  # Hide the first column

        # Define number of columns
        self.minerals_tree['columns'] = ['Element'] + list(self.minerals)

        # Format columns
        for col in self.model.minerals_data.columns:
            self.minerals_tree.column(col, width=60, stretch=False, anchor='w')
            self.minerals_tree.heading(col, text=col)

        # Insert data into the Treeview
        for index, row in self.model.minerals_data.iterrows():
            self.minerals_tree.insert("", tk.END, iid=index, values=[index] + row.tolist())
        self.minerals_tree.column('#1', width=40, stretch=False, anchor='w')

        # Horizontal scrollbar
        hscroll = tk.Scrollbar(pane_tab, orient="horizontal", command=self.minerals_tree.xview)
        hscroll.pack(side="bottom", fill=tk.X)

        self.minerals_tree.configure(xscrollcommand=hscroll.set)
        self.minerals_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.minerals_tree.bind('<<CellUpdated>>', self.update_minerals_data)

        # Extra widgets
        self.extra_widgets()

    def fill_status_bar(self):
        ttk.Separator(self.statusbar, orient='horizontal').pack(side="top", fill="x")

        # Total
        total_label, total_color = total_label_color(self.solid_total)
        tk.Label(self.statusbar, text="Total :").pack(side="left")
        self.label_total = tk.Label(self.statusbar, text=total_label, fg=total_color)
        self.label_total.pack(side="left")

        # About missing minerals or missing oxels
        if len(self.missing_minerals) > 0:
            ttk.Separator(self.statusbar, orient='vertical').pack(side="left", fill="y")
            label_content = "Minerals missing in the dataset: " + str(self.missing_minerals)
            tk.Label(self.statusbar, text=label_content, fg="#f00").pack(side="left")

        if len(self.missing_oxels) > 0:
            ttk.Separator(self.statusbar, orient='vertical').pack(side="left", fill="y")
            label_content2 = "Elements missing in the dataset: " + str(self.missing_oxels)
            tk.Label(self.statusbar, text=label_content2, fg="#f00").pack(side="left")

        for content in self.supplement_status:
            ttk.Separator(self.statusbar, orient='vertical').pack(side="left", fill="y")
            tk.Label(self.statusbar, text=content, ).pack(side="left")

    def toggle_infos(self):
        self.show_infos = not self.show_infos
        self.refresh_plotting()



import math
import numpy as np


def nan_to_none(x, y):
    nx, ny = x, y
    if x is not None and math.isnan(x):
        nx = None
    if y is not None and math.isnan(y):
        ny = None
    return nx, ny


def min_wo_none(x, y):
    x, y = nan_to_none(x, y)
    if (x is None) and (y is None):
        return np.nan
    elif x is None:
        return y
    elif y is None:
        return x
    else:
        return min(x, y)


def gg_max(x, y):
    x, y = nan_to_none(x, y)
    if (x is None) and (y is None):
        return np.nan
    elif x is None:
        return y
    elif y is None:
        return x
    else:
        return max(x, y)


# Function to get min in each colum of two rows
def row_min(old_row, cur_row):
    if not old_row:
        return cur_row

        # Check if same length
    if len(old_row) != len(cur_row):
        raise ValueError('Two rows doesn\' have the same size !')

    new_row = []
    for x, y in zip(old_row, cur_row):
        new_row.append(min_wo_none(x, y))
    return new_row


# Function to get max in each colum of two rows
def row_max(old_row, cur_row):
    if not old_row:
        return cur_row

    # Check if same length
    if len(old_row) != len(cur_row):
        raise ValueError('Two rows doesn\' have the same size !')

    new_row = []
    for x, y in zip(old_row, cur_row):
        new_row.append(gg_max(x, y))

    return new_row


def unique(list1):
    unique_list = []
    for x in list1:
        if x not in unique_list:
            unique_list.append(x)
    return unique_list


def float_or_zero(value):
    try:
        return float(value)
    except (ValueError, TypeError):
        return 0


def linspace(start, stop, num, round_to=3):
    lns = np.linspace(start, stop, num)
    return [round(val, round_to) for val in lns]


def linspace_to_end(b, x, ndigits=5):
    if x <= 1:
        return [b] if x == 1 else []
    step = b / x  # divide into x equal parts, so we skip the 0
    return [round((i + 1) * step, ndigits) for i in range(x)]


def round_ceil_n_digits(value, n):
    multiplier = 10 ** n
    return math.ceil(value * multiplier) / multiplier


def round_floor_n_digits(value, n):
    multiplier = 10 ** n
    return math.floor(value * multiplier) / multiplier


# Get the key from a value in a dictionary
def get_key(dictionary, value):
    key_list = list(dictionary.keys())
    val_list = list(dictionary.values())
    position = val_list.index(value)
    return key_list[position]


def repeat_list(lst, n):
    return [lst] * n


from matplotlib import patches
from matplotlib.colors import to_rgba
from matplotlib.lines import Line2D
from georunes.plot.base import DiagramBase
from georunes.tools.reservoirs import Reservoirs


class PlotLegend(DiagramBase):
    def __init__(self, datasource,
                 show_reservoirs=None,
                 lines_to_break=None,
                 label_order_column='label_order',
                 decor_set=None,
                 title='Legend',
                 alpha_color=0.4, alpha_edge_color=1,
                 **kwargs
                 ):
        DiagramBase.__init__(self, datasource=datasource,
                             title=title, **kwargs)
        self.lines_to_break = lines_to_break
        if show_reservoirs is None:
            show_reservoirs = ["", ],
        self.show_composition = (*show_reservoirs,)
        self.decor_set = decor_set
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.label_order_column = label_order_column

    def is_group_allowed(self, name):
        if (self.exclude_groups is not None) and (name not in self.exclude_groups):
            return True
        return False

    def plot(self):
        if hasattr(self, "window_title"):
            self.fig.canvas.manager.set_window_title(self.window_title)

        self.ax.set_axis_off()
        groups = self.data.groupby(self.group_name)
        list_label_legend = {}
        handles_norm = {}
        res = Reservoirs.get_instance()

        label_order_i = 0
        for name, group in groups:
            if self.is_group_allowed(name):
                if self.label_defined and group[self.label_column].iloc[0]:
                    label = group[self.label_column].iloc[0]
                else:
                    label = name

                # Add group to list of order appearance
                if self.label_order_column in group.keys():
                    list_label_legend[list(group[self.label_order_column])[0]] = label
                else:
                    list_label_legend[label_order_i] = label
                    label_order_i += 1

                # Scatter
                if self.decor_set == "spiderlines":
                    self.ax.plot([], [], linewidth=1, label=label, color=list(group[self.color_column])[0], )

                if self.decor_set == "mixed":
                    color = group[self.color_column].iloc[0]
                    marker = group[self.marker_column].iloc[0]
                    sample_color = to_rgba(color, alpha=self.alpha_color)
                    edge_color = to_rgba(color, alpha=self.alpha_edge_color)
                    self.ax.plot([], [], linewidth=1, label=label, color=list(group[self.color_column])[0],
                                 marker=marker, markeredgecolor=edge_color,
                                 markerfacecolor=sample_color, markersize=self.markersize,
                                 )
                else:
                    sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                    edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                    self.ax.scatter([], [], edgecolors=edge_color, facecolors=sample_color,
                                    marker=list(group[self.marker_column])[0], label=label,
                                    s=self.markersize,
                                    )
                self.fig.subplots_adjust(left=0.1, right=0.9)

        for reference in self.show_composition:
            if reference in res.get_models_list():
                line = Line2D([], [], c=res.get_color(reference), linewidth=1.5, label=res.get_label(reference), )
                handles_norm[res.get_label(reference)] = line

        handles, labels = self.ax.get_legend_handles_labels()
        max_order = max(*[int(x) for x in list_label_legend.keys()])
        ordered_handles = []
        ordered_labels = []
        line = 0
        for i in range(1, max_order + 1):
            if i in list_label_legend.keys():  # If group is not excluded
                idx = labels.index(list_label_legend[i])
                ordered_handles.append(handles[idx])
                ordered_labels.append(labels[idx])
                line += 1
                if self.lines_to_break and line in self.lines_to_break:
                    r = patches.Rectangle((0, 0), 1, 1, fill=False, edgecolor='none', visible=False)
                    ordered_handles.append(r)
                    ordered_labels.append("")

        for key, val in handles_norm.items():
            ordered_handles.append(val)
            ordered_labels.append(key)

        self.ax.legend(ordered_handles, ordered_labels, ncol=self.legend_ncol, loc=self.legend_loc)



from georunes.petromod.modelers.partition import PartitionModeler


class RayleighCrystallization(PartitionModeler):
    def __init__(self, bulk_dist_coeffs):
        PartitionModeler.__init__(self, bulk_dist_coeffs)
        self.phases_labels = {"LIQ": "Instantaneous liquid",
                                 "SOL_INST": "Instantaneous solid",
                                 "SOL_TOT": "Accumulated solid", }
        self.modeler_name = "Fractionated crystallization"
        self.init_name = "Initial liquid composition (F=1)"

    def ratio_liq(self, liq_fract):
        if liq_fract == 0:
            liq_fract = 0.01
        ratio_liq = self.bulk_dist_coeffs.copy()
        for index, row in ratio_liq.items():
            ratio_liq[index] = liq_fract ** (self.bulk_dist_coeffs[index] - 1)
        ratio_liq = ratio_liq.drop(columns=['D', ])

        return ratio_liq

    def ratio_sol(self, liq_fract):
        """
        Calculate the ratio : instantaneous cumulate / initial concentration
        """
        return self.ratio_liq(liq_fract) * self.bulk_dist_coeffs

    def ratio_sol_avg(self, liq_fract):
        """
        Calculate the ratio : total cumulate / initial concentration
        """
        if liq_fract == 1:
            liq_fract = 0.99
        ratio_sol = self.bulk_dist_coeffs.copy()
        for index, row in ratio_sol.items():
            num = 1 - liq_fract ** (self.bulk_dist_coeffs[index])
            ratio_sol[index] = num / (1 - liq_fract)
        return ratio_sol

    def concentration_sol_avg_el(self, el, liq_fract, initial_c0):
        """
        Calculate the ratio for element : instantaneous cumulate / initial concentration
        """
        c_sol_el = initial_c0[el] * self.ratio_sol_avg(liq_fract)[el]
        return c_sol_el

    def get_phase_concentration_func(self, phase):
        if phase == 'LIQ':
            return self.concentration_liq_el
        elif phase == 'SOL_INST':
            return self.concentration_sol_el
        elif phase == 'SOL_TOT':
            return self.concentration_sol_avg_el
        raise Exception("Inapplicable phase : " + str(phase))

    @staticmethod
    def get_phases_labels():
        return {"LIQ": "Liquid", "SOL_INST": "Instantaneous solid", "SOL_TOT": "Accumulated solid"}


class RayleighMelting(PartitionModeler):
    def __init__(self, bulk_dist_coeffs):
        PartitionModeler.__init__(self, bulk_dist_coeffs)
        self.phases_labels = {"LIQ_INST": "Instantaneous liquid",
                                 "LIQ_TOT": "Accumulated liquid",
                                 "SOL": "Solid residue", }
        self.modeler_name = "Rayleigh melting"
        self.init_name = "Initial solid composition (F=0)"

    def ratio_liq(self, liq_fract):
        if liq_fract == 1:
            liq_fract = 0.99
        ratio_liq = self.bulk_dist_coeffs.copy()
        for index, row in ratio_liq.items():
            ratio_liq[index] = (1 / self.bulk_dist_coeffs[index]) * (1 - liq_fract) ** (
                        1 / self.bulk_dist_coeffs[index] - 1)
        return ratio_liq

    def ratio_sol(self, liq_fract):
        return self.ratio_liq(liq_fract) * self.bulk_dist_coeffs

    def ratio_liq_avg(self, liq_fract):
        if liq_fract == 0:
            liq_fract = 0.01
        ratio_liq_avg = self.bulk_dist_coeffs.copy()
        for index, row in ratio_liq_avg.items():
            num = 1 - (1 - liq_fract) ** (1 / self.bulk_dist_coeffs[index])
            ratio_liq_avg[index] = num / liq_fract
        return ratio_liq_avg

    def concentration_liq_avg_el(self, el, liq_fract, initial_conc):
        c_liq_el = initial_conc[el] * self.ratio_liq_avg(liq_fract)[el]
        return c_liq_el

    def get_phase_concentration_func(self, phase):
        if phase == 'LIQ_INST':
            return self.concentration_liq_el
        elif phase == 'SOL':
            return self.concentration_sol_el
        elif phase == 'LIQ_TOT':
            return self.concentration_liq_avg_el
        raise Exception("Inapplicable phase : " + str(phase))



import warnings
import matplotlib.pyplot as plt
from georunes.tools.preprocessing import check_data, data_create_graphic_preset, data_set_graphic_preset
from georunes.tools.filemanager import FileManager
from georunes.tools.warnings import FunctionParameterWarning


class DiagramBase:
    def __init__(self, datasource,
                 sheet=0,
                 no_marker=False,
                 no_title=False, no_legend=False,
                 title="", window_title=None, h_ratio=None,
                 group_name='group', exclude_groups=("",), extra_exclude=None,
                 marker_column='marker', color_column='color', zorder_column=None,
                 supp_group=None,  # Second group for classification data
                 ignore_checkings=False, ignore_checking_markers=False,
                 decor_text_col="k", decor_line_col="k",
                 legend_ncol=1, legend_loc="lower center", legend_in_axs=False,
                 label_column = 'label',
                 arrows=None,
                 padding=None,
                 lang_cfg=None,
                 custom_zorder={},
                 fontsize='medium', title_fs='medium', legend_fs='medium',
                 legend_ms=[50], markersize=None,
                 auto_graphic_preset=True, graphic_preset=None,
                 ):

        filemanager = FileManager.get_instance()
        self.data = filemanager.read_file(datasource, sheet_name=sheet)
        self.check_parameters()  # Verify if some required data are present in file. Can be implemented in child classes
        self.title = title
        self.h_ratio = h_ratio
        self.init_plot()
        if exclude_groups is None:
            self.exclude_groups = ("",)
        else:
            self.exclude_groups = exclude_groups
        if isinstance(extra_exclude, dict):
            for col, value in extra_exclude.items():
                if isinstance(value, (list, tuple)):
                    for e in value:
                        self.data = self.data[self.data[col] != e]
                else:
                    self.data = self.data[self.data[col] != value]

        self.window_title = window_title if window_title else title
        self.no_title = no_title
        self.no_legend = no_legend
        self.group_name = group_name
        self.color_column= color_column
        self.marker_column = marker_column
        self.zorder_column = zorder_column
        self.supp_group = supp_group
        self.datasource = datasource
        self.no_marker = no_marker
        self.custom_zorder = custom_zorder
        self.arrows = arrows
        self.legend_ncol = legend_ncol
        self.legend_loc = legend_loc
        self.legend_in_axs = legend_in_axs
        self.decor_text_col = decor_text_col
        self.decor_line_col = decor_line_col
        self.padding_config = False
        self.init_padding(padding)
        self.lang_cfg = lang_cfg
        self.fontsize = fontsize
        self.title_fs = title_fs
        self.legend_fs = legend_fs
        self.legend_ms = legend_ms
        self.markersize = markersize
        if label_column in self.data.columns :
            self.label_column = label_column
            self.label_defined = True
        else :
            self.label_column = None
            self.label_defined = False
        if not ignore_checkings:
            check_data(self.data, group_name=self.group_name, supp_group=self.supp_group,
                       color_column=self.color_column, marker_column=self.marker_column, zorder_column=self.zorder_column,
                       ignore_checking_markers=ignore_checking_markers)

        if auto_graphic_preset and color_column not in self.data.columns:  # If color is missing, no graphic preset is provided
                if graphic_preset:
                    self.data = data_set_graphic_preset(self.data, graphic_preset, group_name=group_name)
                else:
                    self.data = data_create_graphic_preset(self.data, group_name=group_name)

    def init_padding(self, padding):
        self.padding_left, self.padding_right, self.padding_top, self.padding_bottom = None, None, None, None

        if padding:
            self.padding_config = True
            for side in padding.keys():
                if side == "left":
                    self.padding_left = padding["left"]
                if side == "top":
                    self.padding_top = padding["top"]
                if side == "bottom":
                    self.padding_bottom = padding["bottom"]
                if side == "right":
                    self.padding_right = padding["right"]

    def adjust_padding(self):
        if self.padding_left:
            self.fig.subplots_adjust(left=self.padding_left)
        if self.padding_right:
            self.fig.subplots_adjust(right=self.padding_right)
        if self.padding_top:
            self.fig.subplots_adjust(top=self.padding_top)
        if self.padding_bottom:
            self.fig.subplots_adjust(bottom=self.padding_bottom)

    def init_plot(self):
        if self.h_ratio is None:
            self.fig, self.ax = plt.subplots()
        else:
            if isinstance(self.h_ratio, list):
                self.fig, self.ax = plt.subplots(figsize=self.h_ratio)
            else:
                self.fig, self.ax = plt.subplots(figsize=plt.figaspect(self.h_ratio))

    def is_group_allowed(self, name):
        if type(name) != str and len(
                name) > 1:  # If multiple parameters are chosen to filter group data,
            # take the first one who must be the group
            name = name[0]
        if (self.exclude_groups is not None) and (not str(name) in self.exclude_groups):
            return True
        return False

    def set_xlabel(self, nx):
        self.xlabel = nx

    def set_ylabel(self, ny):
        self.ylabel = ny

    def plot_config(self):
        if hasattr(self, "window_title"):
            self.fig.canvas.manager.set_window_title(self.window_title)
        if self.padding_config:
            self.adjust_padding()

    def marker_size_scaled(self, markersize=None):

        if markersize is None:
            markersize = self.markersize

        if isinstance(markersize, dict):
            chklist = ('var_scale', 'val_max', 'val_min')
            if all(key in markersize.keys() for key in chklist):
                if 'size_max' not in markersize.keys():
                    markersize['size_max'] = None
                if 'size_min' not in markersize.keys():
                    markersize['size_min'] = None
                return True
            else:
                warnings.warn("Key parameters missing for configuration of markersize", FunctionParameterWarning)
        return False

    def check_parameters(self):
        pass

    def save(self, fname, **kwargs):
        self.fig.savefig(fname, **kwargs)

    def save_svg(self, fname, **kwargs):
        self.fig.savefig(fname, **kwargs)



import math

import numpy as np
from scipy import interpolate

sin60 = math.sqrt(3) / 2


def get_spline(x, y, prec=50):
    # get the cumulative distance along the contour
    dist = np.sqrt((x[:-1] - x[1:]) ** 2 + (y[:-1] - y[1:]) ** 2)
    dist_along = np.concatenate(([0], dist.cumsum()))

    # build a spline representation of the contour
    spline, u = interpolate.splprep([x, y], u=dist_along, s=0)

    # resample it at smaller distance intervals
    interp_d = np.linspace(dist_along[0], dist_along[-1], prec)
    interp_x, interp_y = interpolate.splev(interp_d, spline)
    return interp_x, interp_y


def tern_coords_to_bin_coords(rvar, tvar, lvar=None, scale=100):
    xvar = tvar / 2 + rvar
    yvar = tvar * sin60
    return xvar, yvar


def normalize_marker_size(serie, val_max, val_min, size_max, size_min):
    if size_max is None:
        size_max = 100
    if size_min is None:
        size_min = 1
    serie = np.clip(serie, a_min=val_min, a_max=None)
    serie = np.clip(serie, a_min=None, a_max=val_max)
    a = (size_max - size_min) / (val_max - val_min)
    b = size_max - a * val_max
    size_serie = serie * a + b
    return size_serie


def is_in_canvas(x, y, xlim, ylim):
    # Check if element is in canvas
    return (x > xlim[0]) and (x < xlim[1]) and (y > ylim[0]) and (y < ylim[1])



import numpy as np
from numpy import linalg
from random import random


def is_in_bounds(partition, max_minerals_prop, min_minerals_prop):
    if any(partition[i] > max_minerals_prop[i] or partition[i] < min_minerals_prop[i] for i in range(len(partition))):
        return False
    return True


def random_part_in_hypercube(current_part, semiedge, max_minerals_prop=None, min_minerals_prop=None, total=1,
                             unfillable_partitions_allowed=True, max_iter=100, verbose=0):
    nb_minerals = len(current_part)
    if not min_minerals_prop:
        min_minerals_prop = [0] * nb_minerals
    if not max_minerals_prop:
        max_minerals_prop = [1] * nb_minerals
    if len(max_minerals_prop) != nb_minerals or len(min_minerals_prop) != nb_minerals:
        raise Exception("Wrong dimensions for max or min mineral compositions")

    it = 0
    prop = random()  # A random factor to allow smaller steps during the research of new values
    while it < max_iter:
        it += 1
        if unfillable_partitions_allowed :
            dec = np.random.uniform(-semiedge * prop, semiedge * prop, nb_minerals)
        else :
            dec = np.random.uniform(-semiedge * prop, semiedge * prop, nb_minerals - 1)
            dec = np.append(dec, total - sum(current_part) - sum(dec))  # To make sure that sum(new_partition) = total
        new_partition = [current_part[i] + dec[i] for i in range(nb_minerals)]
        if is_in_bounds(new_partition, max_minerals_prop, min_minerals_prop):
            # The new composition is found
            return new_partition

        if verbose > 1 and it == max_iter:
            print("WARNING : Maximum iteration reached in the research of a random composition.")

    return current_part


def random_part_with_bounds(nb_minerals, max_minerals_prop=None, min_minerals_prop=None, total=1,
                            unfillable_partitions_allowed=False, verbose=0, ):
    if not min_minerals_prop:
        min_minerals_prop = [0] * nb_minerals
    if not max_minerals_prop:
        max_minerals_prop = [1] * nb_minerals

    if len(max_minerals_prop) != nb_minerals or len(min_minerals_prop) != nb_minerals:
        raise Exception("Wrong dimensions for max or min mineral compositions")

    # Loop
    found = False
    k = 0
    while not found:
        k += 1
        new_partition = []
        for i in range(nb_minerals):
            new_partition.append(min_minerals_prop[i] + (max_minerals_prop[i] - min_minerals_prop[i]) * random())
        new_total = sum(new_partition)
        found = True
        if new_total > total:
            new_partition = [i * total / new_total for i in new_partition]
            # Check if minima conditions are respected
            for j in range(nb_minerals):
                if new_partition[j] < min_minerals_prop[j]:
                    found = False
                    break
        elif not unfillable_partitions_allowed:
            new_partition = [i * total / new_total for i in new_partition]
            for j in range(nb_minerals):
                if new_partition[j] > max_minerals_prop[j]:
                    found = False
                    break
        else:  # if unfillable partitions are authorized, nothing to do.
            # From its definition, new_partition respects the maxima and minima conditions
            pass
        if k > 1000:
            raise Exception("The algorithm struggles to find a coherent random partition. Please check the entry data.")
    if verbose > 2:
        print("Partition found after", k, "iterations.")

    return new_partition


def random_part(nb_minerals, verbose=0):
    return random_part_with_bounds(nb_minerals, verbose=verbose)


class BaseOptimizer:
    def __init__(self, verbose=None):
        if isinstance(verbose, (int, bool)):
            self.verbose = int(verbose)
        else:
            self.verbose = 0

    def deviation(self, part_a, part_b):
        if self.dist_func in ("euclidian", "L2-norm"):
            dist = np.linalg.norm(part_a - part_b)
        elif self.dist_func == "max":
            dist = np.linalg.norm(part_a - part_b, ord=np.inf)
        elif self.dist_func == "min":
            dist = np.linalg.norm(part_a - part_b, ord=-np.inf)
        elif self.dist_func == "MAE":  # Mean absolute error
            dist = np.absolute(part_a - part_b).mean()
        elif self.dist_func == "MSE":  # Mean square error
            dist = np.square(part_a - part_b).mean()
        elif self.dist_func == "RMSE":  # Root mean square error
            dist = np.sqrt(np.square(part_a - part_b).mean())
        elif self.dist_func == "SMAPE":  # Symmetric mean absolute percentage error
            sum_abs_a_b = np.abs(part_a) + np.abs(part_b)
            terms = np.divide(2 * np.abs(part_a - part_b), sum_abs_a_b, out=np.zeros_like(part_a),
                              where=sum_abs_a_b != 0)  # Put division to 0 when A = B = 0
            dist = 100 * terms.mean()
        else:
            raise Exception("Unknown parameter for distance function.")
        return dist


class Optimizer(BaseOptimizer):
    def __init__(self, **kwargs):
        BaseOptimizer.__init__(self, **kwargs)
        # self.norm defined in children classes

    def prepare_data(self, raw_data, skip_cols, raw_minerals_data, ignore_oxides):
        raw_data = raw_data.fillna(0)
        if 'Sum' in raw_data.keys():
            raw_data = raw_data.rename(columns={'Sum': 'Total'})
        if 'Total' in raw_data.keys():
            self.init_total = raw_data["Total"]
        else:
            raise Exception("Total column not found.")

        if ignore_oxides:
            el_ox_to_drop = [*ignore_oxides, 'Total']
        else:
            el_ox_to_drop = ['Total', ]

        for ox in el_ox_to_drop:
            if ox in raw_minerals_data.keys():
                raw_minerals_data = raw_minerals_data.drop(columns=ox)
            if ox in raw_data.keys():
                raw_data = raw_data.drop(columns=ox)

        raw_minerals_data = raw_minerals_data.fillna(0).infer_objects(copy=False)
        raw_minerals_data = raw_minerals_data.set_index(raw_minerals_data.keys()[0])
        self.data = raw_data.iloc[:, skip_cols:].copy()
        self.list_bulk_ox = self.data.keys().tolist()
        raw_minerals_data = raw_minerals_data[[*list(self.list_bulk_ox)]]  # Order oxides as in source
        self.minerals_data = raw_minerals_data.transpose()
        self.list_minerals = self.minerals_data.keys().tolist()
        self.nb_minerals = len(self.list_minerals)

    def compute(self, raw_data, skip_cols, raw_minerals_data, to_round=4, ignore_oxides=None, ratios=None, **kwargs):
        if self.verbose:
            print(self.notif)

        if ratios:
            notif_cfg = ""
            rmd_t = raw_minerals_data.transpose()
            rmd_t.columns = rmd_t.iloc[0]
            df_t = rmd_t[1:].copy()
            for sol, cfg in ratios.items():
                if len(cfg) != 2:
                    print("WARNING : Incorrect configuration for the solution :", sol)
                elif len(cfg[0]) != len(cfg[1]):
                    print("WARNING : Missing data in the solution :", sol)
                else:
                    mins, parts = cfg[0], cfg[1]
                    if notif_cfg:
                        notif_cfg += " / " + str([mins, parts, ])
                    else:
                        notif_cfg += str([mins, parts, ])

                    # Calculation of mineral solutions and replacement in minerals_data
                    tot = sum(parts)
                    for miner, pa in zip(mins, parts):
                        df_t[miner] = df_t[miner] * pa / tot
                    df_t[sol] = df_t[mins].sum(axis=1)
                    df_t = df_t.drop(columns=mins)

            if self.verbose and len(notif_cfg) > 0:
                print("Considered mineral solutions :", notif_cfg)

                # Reset dataframe as raw table
                sub_data = df_t.transpose()
                sub_data.reset_index(drop=False, inplace=True)

                # Compute partitions
                partitions, suppl = self._compute(raw_data, skip_cols, sub_data, to_round=to_round,
                                                  ignore_oxides=ignore_oxides, **kwargs)

                # Component reconstitution
                for sol, cfg in ratios.items():
                    mins, parts = cfg[0], cfg[1]
                    tot = sum(parts)
                    suppl[sol] = partitions[sol]
                    for miner, pa in zip(mins, parts):
                        partitions[miner] = partitions[sol] * pa / tot
                    partitions = partitions.drop(columns=sol, axis=1)
                # Move Total to the last column
                cols = partitions.columns.tolist()
                cols.remove("Total")
                cols.append("Total")
                partitions = partitions[cols]
            else:
                partitions, suppl = self._compute(raw_data, skip_cols, raw_minerals_data, to_round=to_round,
                                                  ignore_oxides=ignore_oxides, **kwargs)

        else:
            partitions, suppl = self._compute(raw_data, skip_cols, raw_minerals_data, to_round=to_round,
                                              ignore_oxides=ignore_oxides, **kwargs)

        if self.verbose:
            self.show_results(partitions, suppl)

        return partitions, suppl

    @staticmethod
    def show_results(partitions, suppl):
        print(">>> Final compositions (wt %)")
        print(partitions.to_string())
        print(">>> Supplementary data")
        print(suppl.to_string())

    def set_verbose(self, verbose):
        self.verbose = verbose
        print("verbose now is", self.verbose)



import warnings


class GeoRunesWarning(Warning):
    """Base warning class for GeoRunes package."""
    pass


class DataFormatWarning(GeoRunesWarning):
    """Raised when input data format is not supported or inconsistent."""
    pass


class DataIntegrityWarning(GeoRunesWarning):
    """Raised when data quality issues are detected (e.g., NaNs, outliers)."""
    pass


class UnitsWarning(GeoRunesWarning):
    """Raised when inconsistent or unexpected units are detected."""
    pass


class FunctionParameterWarning(GeoRunesWarning):
    """Raised when an invalid parameter value is provided."""
    pass


class InsufficientDataWarning(GeoRunesWarning):
    """Raised when input data is insufficient for reliable modal estimation."""
    pass


class ConvergenceWarning(GeoRunesWarning):
    """Raised when optimization methods fail to converge."""
    pass


class PlotDataWarning(GeoRunesWarning):
    """Warning related to issues in data plotting."""
    pass



import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
from georunes.petromod.viewer.tk.fract_viewer import FractViewerTk
from georunes.tools.reservoirs import get_reservoir_norm


class SpiderViewerTk(FractViewerTk):
    def __init__(self, petro_model, data_coeffs, initial_conc, norm="CI", cmap='plasma_r', *args, **kwargs):
        self.norm = get_reservoir_norm(norm)
        self.line = None # For ax.semilogy
        self.cmap = plt.get_cmap(cmap)
        FractViewerTk.__init__(self, petro_model, data_coeffs, initial_conc, *args, **kwargs)

    def set_supplement_status(self):
        self.supplement_status = []
        norm_status = "Norm : " + str(self.norm["reservoir"])
        self.supplement_status.append(norm_status)

    def draw(self, element=None):
        list_elements = [el for el in self.model.list_elements() if el in self.initial_conc_elts]
        concentration_func = self.model.get_phase_concentration_func(self.selected_phase)
        for f in self.liq_fract_values:
            values = [concentration_func(el, f, self.initial_conc) / self.norm[el] for el in list_elements]
            self.ax.semilogy(list_elements, values, label='F={}'.format(f), color=self.cmap(f))
        self.ax.set_ylabel('Normalized concentrations')
        self.ax.grid(axis='y', which='both', linestyle='-', color='#e8e8e8')
        if self.show_legend:
            cbar = plt.colorbar(ScalarMappable(cmap=self.cmap, norm=Normalize(0, 1)), ax=self.ax)
            cbar.set_label('Liquid fraction (F)')
        self.canvas.draw()

    def fill_frame(self):
        FractViewerTk.fill_frame(self)



import math
import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba

from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer
from georunes.plot.piper.piper_axe_subplot import PiperAxesSubplot
from georunes.tools.plotting import normalize_marker_size


class DiagramPiper(DiagramBase, LegendDrawer):
    def __init__(self, datasource,
                 padding=None,
                 marker='', annotation=None,
                 alpha_color=0.4,  alpha_edge_color=0.8,
                 legend_ncol=4,
                 delta=0.1,
                 no_ticks=False, no_ticks_label=False,
                 ticks_number=6, ticks_fontsize="small", ticks_clockwise=True, ticks_offset=0.03,
                 nb_grid_separators=9, show_gridlines=True,
                 show_axis=False,
                 h_ratio=None,
                 **kwargs
                 ):
        self.delta = delta
        if padding is None:
            padding = {"left": 0.06, "right": 0.98}
        self.fig = None
        self.ax = None
        self.pax = None
        DiagramBase.__init__(self, datasource=datasource,  h_ratio=h_ratio, legend_ncol=legend_ncol, padding=padding,
                             **kwargs)
        self.annotation = annotation
        self.marker = marker
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.no_ticks = no_ticks
        self.no_ticks_label = no_ticks_label
        self.ticks_number = ticks_number
        self.ticks_fontsize = ticks_fontsize
        self.ticks_clockwise = ticks_clockwise
        self.ticks_offset = ticks_offset
        self.show_gridlines = show_gridlines
        self.nb_grid_separators = nb_grid_separators
        self.show_axis = show_axis

    def init_plot(self):
        ratio_parameter = math.sqrt(3) / 2
        if self.h_ratio:
            ratio_parameter = self.h_ratio
        self.fig, self.ax = plt.subplots(figsize=plt.figaspect(ratio_parameter))  # Ratio figure size
        self.pax = PiperAxesSubplot(ax=self.ax, delta=self.delta)

    def plot_config(self):
        DiagramBase.plot_config(self)

        if self.show_gridlines:
            self.pax.gridlines(color="black", num_seps=self.nb_grid_separators, linewidth=0.5)
        self.pax.boundary(linewidth=1.0)

        if not self.no_title:
            if self.title == "":
                self.title = 'Piper diagram'
            self.pax.set_title(self.title)

        if not self.no_ticks:  # Show ticks
            self.pax.ticks(linewidth=1, num_ticks=self.ticks_number, clockwise=self.ticks_clockwise,
                           offset=self.ticks_offset, fontsize=self.ticks_fontsize)

        if not self.no_ticks_label:  # Show labels
            self.pax.set_labels_visible()

    def set_decorations(self, ):
        self.pax.set_xlim(-0.1, 1.1)
        self.pax.set_ylim(-0.08, 0.94)
        if not self.show_axis:
            self.pax.set_axis_off()

    def plot(self):
        self.plot_config()
        self.set_decorations()

        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):

                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = None
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                # Convert points to tuples
                points = [(group["Ca"].get(i), group["Mg"].get(i), group["Na"].get(i), group["K"].get(i),
                           group["HCO3"].get(i), group["CO3"].get(i), group["SO4"].get(i), group["Cl"].get(i))
                          for i, sample in group["Ca"].items()]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.pax.scatter(points, edgecolors=edge_color,
                                 marker=marker, label=label, facecolors=sample_color, s=size,
                                 zorder=zorder)

        self.plot_legend()



from functools import partial
import numpy as np
from matplotlib import pyplot
from georunes.plot.piper import lines
from georunes.plot.piper.helpers import project_left_triangle, project_right_triangle, project_diamond
from georunes.plot.piper.plotting import scatter, plot
from georunes.tools.language import format_chemical_formula as _fml


def figure(ax=None, delta=0.2):
    """
    Wraps a PiperAxesSubplot
    """
    piper_ax = PiperAxesSubplot(ax=ax, delta=delta)
    return piper_ax.get_figure(), piper_ax


def mpl_redraw_callback(event, pax):
    """
    Callback to properly rotate and redraw text labels when the plot is drawn
    or resized.

    Parameters
    ----------
    event: a matplotlib event
        either 'resize_event' or 'draw_event'
    pax: PiperAxesSubplot
         the PiperAxesSubplot
    """
    pax._redraw_labels()


class PiperAxesSubplot(object):
    """
    AxesSubplot for Piper diagram
    """

    def __init__(self, ax=None, delta=0.1, labels_fontsize=8):

        if ax:
            self.ax = ax
        else:
            _, self.ax = pyplot.subplots()

        # Container for the axis labels supplied by the user
        self._labels_fontsize = labels_fontsize
        self._ticks = dict()
        # Container for the redrawing of labels
        self._to_remove = []
        self._connect_callbacks()
        self._delta = delta
        self._hide_labels = True

    def _connect_callbacks(self):
        """Connect resize matplotlib callbacks."""
        fig = self.get_figure()
        callback = partial(mpl_redraw_callback, pax=self)
        event_names = ('resize_event', 'draw_event')
        for event_name in event_names:
            fig.canvas.mpl_connect(event_name, callback)

    def __repr__(self):
        return "PiperAxesSubplot: %s" % self.ax.__hash__()

    def get_axes(self):
        """Returns the underlying PiperAxesSubplot object."""
        return self.ax

    def get_figure(self):
        """Return the underlying matplotlib figure object."""
        ax = self.get_axes()
        return ax.get_figure()

    def set_axis_off(self):
        ax = self.get_axes()
        ax.set_axis_off()

    def set_title(self, title, **kwargs):
        """Sets the title on the underlying PiperAxesSubplot."""
        ax = self.get_axes()
        ax.set_title(title, **kwargs)

    def set_labels_visible(self):
        self._hide_labels = False

    def boundary(self, axes_colors=None, **kwargs):
        ax = self.get_axes()
        lines.boundary(ax=ax, axes_colors=axes_colors, delta=self._delta, **kwargs)

    def gridlines(self, num_seps=9, **kwargs):
        ax = self.get_axes()
        lines.gridlines(num_seps=num_seps, ax=ax, delta=self._delta, **kwargs)

    def close(self):
        fig = self.get_figure()
        pyplot.close(fig)

    def legend(self, *args, **kwargs):
        ax = self.get_axes()
        ax.legend(*args, **kwargs)

    def savefig(self, filename,  **kwargs):
        self._redraw_labels()
        fig = self.get_figure()
        fig.savefig(filename, **kwargs)

    def show(self):
        self._redraw_labels()
        pyplot.show()

    def ticks(self, ticks=None, num_ticks=6, locations=None, clockwise=False, axes_colors=None, tick_formats=None,
              **kwargs):
        if axes_colors is None:
            axes_colors = "black"
        ax = self.get_axes()
        delta = self._delta
        lines.ticks(ax, delta=delta, list_ticks=ticks, num_ticks=num_ticks, locations=locations,
                    clockwise=clockwise, axes_colors=axes_colors, tick_formats=tick_formats,
                    **kwargs)

    def set_xlim(self, xmin, xmax):
        ax = self.get_axes()
        return ax.set_xlim(xmin, xmax)

    def set_ylim(self, ymin, ymax):
        ax = self.get_axes()
        return ax.set_ylim(ymin, ymax)

    # Redrawing and resizing
    def _redraw_labels(self):
        """Redraw axis labels, typically after draw or resize events."""
        if not self._hide_labels:
            ax = self.get_axes()

            # Remove any previous labels
            for mpl_object in self._to_remove:
                mpl_object.remove()
            self._to_remove = []

            # Redraw the labels with the appropriate angles
            label_data = [
                ["Ca", project_left_triangle((1, 0, 0), delta=self._delta, drift=(0, -0.08)), "center", 0],
                ["Mg", project_left_triangle((0, 1, 0), delta=self._delta, drift=(-0.01, 0.05)), "center", 0],
                ["Na+K", project_left_triangle((0, 0, 1), delta=self._delta, drift=(0, -0.08)), "center", 0],
                [_fml("HCO3+CO3"), project_right_triangle((1, 0, 0),
                                                          delta=self._delta, drift=(0.03, -0.08)), "center", 0],
                [_fml("SO4"), project_right_triangle((0, 1, 0), delta=self._delta, drift=(0, 0.05)), "center", 0],
                ["Cl", project_right_triangle((0, 0, 1), delta=self._delta, drift=(0, -0.08)), "center", 0],
                [_fml("SO4+Cl"), project_diamond((0, 0.5), delta=self._delta, drift=(-0.07, 0)), "center", 60],
                ["Ca+Mg", project_diamond((0.5, 1), delta=self._delta, drift=(0.07, 0)), "center", -60],
            ]

            for label, position, halign, rotation in label_data:
                x, y = position[0], position[1]
                pos_a = np.array([x, y])
                new_rotation = ax.transData.transform_angles(np.array((rotation,)), pos_a.reshape((1, 2)))[0]
                text = ax.text(x, y, label, horizontalalignment=halign, rotation=new_rotation, transform=ax.transData,
                               fontsize=self._labels_fontsize, )
                self._to_remove.append(text)

    def scatter(self, points, **kwargs):
        delta = self._delta
        ax = self.get_axes()
        plot_ = scatter(points, ax=ax, delta=delta, **kwargs)
        return plot_

    def plot(self, points, **kwargs):
        ax = self.get_axes()
        plot(points, delta=self._delta, ax=ax, **kwargs)



import warnings
import numpy as np
from pandas import Series
from georunes.tools.chemistry import ratio_el_to_ox, molar_mass, el_molar_mass
from georunes.modmin.optim.base import BaseOptimizer
from georunes.tools.warnings import DataIntegrityWarning

list_major_ox = ['SiO2', 'Na2O', 'K2O', 'CaO', 'MnO', 'FeO', 'Fe2O3', 'MgO', 'TiO2', 'Al2O3', 'P2O5']
list_minor_el = ['F', 'Cl', 'S', 'Ni', 'Co', 'Ba', 'Sr', 'Rb', 'Cs', 'Li', 'Zr', 'Cr', 'V']
list_ox_from_minor_el = ['SO3', 'NiO', 'CoO', 'BaO', 'SrO', 'Rb2O', 'Cs2O', 'Li2O', 'ZrO2', 'Cr2O3', 'V2O3']
list_all_ox_el = [*list_major_ox, 'CO2', 'F', 'Cl', 'S', *list_ox_from_minor_el]
final_list_min = ['Q', 'Or', 'Ab', 'An', 'C', 'Wo', 'Cs', 'Tn', 'Pf', 'Cc', 'Mt', 'Hm', 'Il', 'Ru',
                  'Ne', 'Th', 'Nc', 'Ac', 'Ns', 'Lc', 'Kp', 'Ks', 'Z', 'Hl', 'Fr', 'Pr', 'Cm']

oxide_corresp = {
    'Ni': 'NiO',
    'Co': 'CoO',
    'Ba': 'BaO',
    'Sr': 'SrO',
    'Rb': 'Rb2O',
    'Cs': 'Cs2O',
    'Li': 'Li2O',
    'Zr': 'ZrO2',
    'Cr': 'Cr2O3',
    'V': 'V2O3'
}

_mol_w = dict()
for _oxel in list_all_ox_el:
    if _oxel in molar_mass:
        _mol_w[_oxel] = molar_mass[_oxel]
    elif _oxel in el_molar_mass:
        _mol_w[_oxel] = el_molar_mass[_oxel]
for _oxel in ['O', 'Cl', 'S']:
    _mol_w[_oxel] = el_molar_mass[_oxel]

# Algorithm updated from Verma et al., 2003
# Verma, S.P., Torres-Alvarado, I.S. and Velasco-Tapia, F., 2003. A revised CIPW norm. Swiss Bulletin of Mineralogy
# and Petrology, 83(2), pp.197-216.

class CIPWNorm(BaseOptimizer):
    def __init__(self, **kwargs):
        BaseOptimizer.__init__(self, **kwargs)
        self.notif = ">>>>>> CIPW norm"

    def compute(self, raw_data, skip_cols, normalize_entry=False, minor_included=False, to_round=4,
                co2_cancrinite=False, co2_calcite=False):

        # 1 / Preparing data
        if self.verbose > 1: print("Step 1 - Prepare data")

        co2_cancrinite = float(co2_cancrinite)
        co2_calcite = float(co2_calcite)
        data = raw_data.iloc[:, :skip_cols].copy()
        raw_data_keys = raw_data.iloc[:, skip_cols:].keys()

        for oxel in raw_data_keys:
            if oxel in [*list_all_ox_el, *list_minor_el]:
                data[oxel] = raw_data[oxel]

        if 'Total' in raw_data_keys or 'Sum' in raw_data_keys:
            data['Total'] = raw_data['Total']
        else:
            warnings.warn("Total column not found.", DataIntegrityWarning)

        data_keys = data.keys()
        filtered_columns = [oxel for oxel in raw_data_keys if oxel not in data_keys]
        if self.verbose: print('Ignored columns :', filtered_columns)

        for ox in list_major_ox:
            if ox not in data_keys:
                data[ox] = 0.

        n_phase = data.iloc[:, :skip_cols].copy()
        next_step = Series(np.array([""] * len(data.index)))
        si_deff = data.iloc[:, :skip_cols].copy()
        prop = data.iloc[:, :skip_cols].copy()
        free = data.iloc[:, :skip_cols].copy()
        partitions = data.iloc[:, :skip_cols].copy()
        suppl = data.iloc[:, :skip_cols].copy()

        # Initializing columns for SiO2 deficiency calculation
        si_deff['Y'] = 0.
        si_deff['D'] = 0.

        # 2 / CO2 handling options
        if self.verbose > 1: print("Step 2 - CO2 handling options")

        if minor_included:
            if co2_cancrinite + co2_calcite != 1 and not (co2_cancrinite == co2_calcite == 0):
                raise Exception('Check the proportions of co2_cancrinite and co2_calcite.')
            if self.verbose > 1:
                print('CO2 proportion in calcite : ', co2_calcite, ' --- CO2 proportion in cancrinite:', co2_cancrinite)

        # 3 / Conversion of trace oxide/element units
        if self.verbose > 1: print("Step 3 - Conversion of trace oxide/element units")

        if minor_included:
            data_keys = data.keys()
            for oxel in data_keys:
                if oxel in list_minor_el and oxel not in ['F', 'Cl', 'S']:
                    oxide = oxide_corresp[oxel]
                    data[oxide] = data[oxel] / (ratio_el_to_ox[oxide] * 10000)
                    data = data.drop(columns=oxel)
                elif oxel in ['F', 'Cl', 'S']:
                    data[oxel] = data[oxel] / 10000

            # If some minor oxide/element does not exist in data, add the corresponding oxide with zero values for
            # treatment. Also add other required oxels that are missing.

            data_keys = data.keys()
            for ox in list_ox_from_minor_el:
                if ox not in data_keys:
                    data[ox] = 0.
            if 'F' not in data_keys:
                data['F'] = 0.

        # 4-5 / Adjust components to 100%
        if self.verbose > 1: print("Step 4, 5 - Adjust components to 100%")

        data_keys = data.keys()
        total_calculated = data.iloc[:, skip_cols:].sum(axis=1) - data['Total']
        if self.verbose > 1:
            print(">>> Total calculated")
            print(total_calculated.to_string())

        if normalize_entry:
            oxel_to_sum = list_major_ox if minor_included else list_major_ox
            for oxel in oxel_to_sum:
                if oxel in data_keys:
                    data[oxel] = 100 * data[oxel] / total_calculated
            total_normalized = data.iloc[:, skip_cols:].sum(axis=1) - data['Total']

            if self.verbose > 1:
                print(">>> Total normalized")
                print(total_normalized.to_string())
        else:
            total_normalized = total_calculated

        # Calculation of some petrogenetical parameters
        suppl['pp_FeOt/MgO'] = ((2 * _mol_w['FeO'] / _mol_w['Fe2O3']) * data['Fe2O3'] + data['FeO']) / data['MgO']
        suppl['pp_ratio_K2O_Na2O'] = data['K2O'] / data['Na2O']
        suppl['pp_SI'] = 100 * data['MgO'] / (data['MgO'] + data['FeO'] + data['Fe2O3'] + data['Na2O'] + data['K2O'])
        suppl['pp_Mg#'] = 100 * (data['MgO'] / _mol_w['MgO']) / (
                data['MgO'] / _mol_w['MgO'] + data['FeO'] / _mol_w['FeO'])

        for i in data.index:
            if data.loc[i, 'SiO2'] > 5 and 1 < suppl.loc[i, 'pp_ratio_K2O_Na2O'] < 2.5:
                suppl.loc[i, 'pp_AR'] = (data.loc[i, 'Al2O3'] + data.loc[i, 'CaO'] + 2 * data.loc[i, 'Na2O']) / (
                        data.loc[i, 'Al2O3'] + data.loc[i, 'CaO'] - 2 * data.loc[i, 'Na2O'])
            else:
                suppl.loc[i, 'pp_AR'] = (data.loc[i, 'Al2O3'] + data.loc[i, 'CaO'] + data.loc[i, 'Na2O'] + data.loc[
                    i, 'K2O']) / (data.loc[i, 'Al2O3'] + data.loc[i, 'CaO'] - data.loc[i, 'Na2O'] - data.loc[i, 'K2O'])

        # 6 / Mole computations
        if self.verbose > 1: print("Step 6 - Mole computations")

        data_n = data.copy().drop(['Total'], axis=1).fillna(0)

        if self.verbose > 1:
            print(">>> Initial compositions (wt %)")
            print(data_n.to_string())

        for oxel in data_n.keys():
            if oxel in list_all_ox_el:
                data_n[oxel] = data_n[oxel] / _mol_w[oxel]

        if self.verbose:
            print(">>> Initial molar concentrations (mol)")
            print(data_n.to_string())

        # 7-8 / Minor oxides combination and oxide molecular weight computations
        if self.verbose > 1: print(
            "Step 7, 8 - Minor oxides combination and oxide molecular weight computations")

        if minor_included:
            FeO_corr = data_n['FeO'] + data_n['MnO'] + data_n['NiO'] + data_n['CoO']
            CaO_corr = data_n['CaO'] + data_n['BaO'] + data_n['SrO']
            K2O_corr = data_n['K2O'] + data_n['Rb2O'] + data_n['Cs2O']
            Na2O_corr = data_n['Na2O'] + data_n['Li2O']
            Cr2O3_corr = data_n['Cr2O3'] + data_n['V2O3']
            xFeO = data_n['FeO'] / FeO_corr
            xMnO = data_n['MnO'] / FeO_corr
            xNiO = data_n['NiO'] / FeO_corr
            xCoO = data_n['CoO'] / FeO_corr
            xCaO = data_n['CaO'] / CaO_corr
            xBaO = data_n['BaO'] / CaO_corr
            xSrO = data_n['SrO'] / CaO_corr
            xK2O = data_n['K2O'] / K2O_corr
            xRb2O = data_n['Rb2O'] / K2O_corr
            xCs2O = data_n['Cs2O'] / K2O_corr
            xNa2O = data_n['Na2O'] / Na2O_corr
            xLi2O = data_n['Li2O'] / Na2O_corr
            xCr2O3 = data_n['Cr2O3'] / Cr2O3_corr
            xV2O3 = data_n['V2O3'] / Cr2O3_corr

            data_n['FeO'] = data_n['FeO'] + data_n['MnO'] + data_n['NiO'] + data_n['CoO']
            data_n['CaO'] = data_n['CaO'] + data_n['BaO'] + data_n['SrO']
            data_n['K2O'] = data_n['K2O'] + data_n['Rb2O'] + data_n['Cs2O']
            data_n['Na2O'] = data_n['Na2O'] + data_n['Li2O']
            data_n['Cr2O3'] = data_n['Cr2O3'] + data_n['V2O3']

            # Other oxides to zero
            for ox in ['MnO', 'NiO', 'CoO', 'BaO', 'SrO', 'Rb2O', 'Cs2O', 'Li2O', 'V2O3']:
                data_n[ox] = 0.

        else:
            FeO_corr = data_n['MnO'] + data_n['FeO']
            xFeO = data_n['FeO'] / FeO_corr
            xMnO = data_n['MnO'] / FeO_corr
            data_n['FeO'] = data_n['FeO'] + data_n['MnO']
            data_n['MnO'] = 0.

        if self.verbose:
            print(">>> Corrected molar concentrations (mol)")
            print(data_n.to_string())

        # Molecular weight correction
        corr_mol_w = _mol_w.copy()  # Each value will be a number of a list of number (for FeO for example)
        if minor_included:
            corr_mol_w['FeO'] = xFeO * _mol_w['FeO'] + xMnO * _mol_w['MnO'] + \
                                xNiO * _mol_w['NiO'] + xCoO * _mol_w['CoO']
            corr_mol_w['CaO'] = xCaO * _mol_w['CaO'] + xBaO * _mol_w['BaO'] + xSrO * _mol_w['SrO']
            corr_mol_w['K2O'] = xK2O * _mol_w['K2O'] + xRb2O * _mol_w['Cs2O'] + xCs2O * _mol_w['Cs2O']
            corr_mol_w['Na2O'] = xNa2O * _mol_w['Na2O'] + xLi2O * _mol_w['Li2O']
            corr_mol_w['Cr2O3'] = xCr2O3 * _mol_w['Cr2O3'] + xV2O3 * _mol_w['V2O3']

            corr_mol_w['FeO'] = corr_mol_w['FeO'].fillna(0)
            corr_mol_w['CaO'] = corr_mol_w['CaO'].fillna(0)
            corr_mol_w['K2O'] = corr_mol_w['K2O'].fillna(0)
            corr_mol_w['Na2O'] = corr_mol_w['Na2O'].fillna(0)
            corr_mol_w['Cr2O3'] = corr_mol_w['Cr2O3'].fillna(0)

        else:
            corr_mol_w['FeO'] = xFeO * _mol_w['FeO'] + xMnO * _mol_w['MnO']
            corr_mol_w['FeO'] = corr_mol_w['FeO'].fillna(0)

        # 9-10 / Correction of normative mineral molecular weights
        if self.verbose > 1: print("Step 9, 10 - Correction of normative mineral molecular weights")

        mol_w_min = {
            'Ab': corr_mol_w['Na2O'] + corr_mol_w['Al2O3'] + 6 * corr_mol_w['SiO2'],
            'Ne': corr_mol_w['Na2O'] + corr_mol_w['Al2O3'] + 2 * corr_mol_w['SiO2'],
            'Th': corr_mol_w['Na2O'] + corr_mol_w['SO3'],
            'Nc': corr_mol_w['Na2O'] + corr_mol_w['CO2'],
            'An': corr_mol_w['CaO'] + corr_mol_w['Al2O3'] + 2 * corr_mol_w['SiO2'],
            'Di-mg': corr_mol_w['CaO'] + corr_mol_w['MgO'] + 2 * corr_mol_w['SiO2'],
            'Di-fe': corr_mol_w['CaO'] + corr_mol_w['FeO'] + 2 * corr_mol_w['SiO2'],
            'Ac': corr_mol_w['Na2O'] + corr_mol_w['Fe2O3'] + 4 * corr_mol_w['SiO2'],
            'Ns': corr_mol_w['Na2O'] + corr_mol_w['SiO2'],
            'Or': corr_mol_w['K2O'] + corr_mol_w['Al2O3'] + 6 * corr_mol_w['SiO2'],
            'Lc': corr_mol_w['K2O'] + corr_mol_w['Al2O3'] + 4 * corr_mol_w['SiO2'],
            'Kp': corr_mol_w['K2O'] + corr_mol_w['Al2O3'] + 2 * corr_mol_w['SiO2'],
            'Ks': corr_mol_w['K2O'] + corr_mol_w['SiO2'],
            'Wo': corr_mol_w['CaO'] + corr_mol_w['SiO2'],
            'Cs': 2 * corr_mol_w['CaO'] + corr_mol_w['SiO2'],
            'Mt': corr_mol_w['FeO'] + corr_mol_w['Fe2O3'],
            'Il': corr_mol_w['FeO'] + corr_mol_w['TiO2'],
            'Tn': corr_mol_w['CaO'] + corr_mol_w['TiO2'] + corr_mol_w['SiO2'],
            'Pf': corr_mol_w['CaO'] + corr_mol_w['TiO2'],
            'Cc': corr_mol_w['CaO'] + corr_mol_w['CO2'],
            'Hy-fe': corr_mol_w['FeO'] + corr_mol_w['SiO2'],
            'Ol-fe': 2 * corr_mol_w['FeO'] + corr_mol_w['SiO2'],
            'Q': corr_mol_w['SiO2'],
            'C': corr_mol_w['Al2O3'],
            'Z': corr_mol_w['SiO2'] + corr_mol_w['ZrO2'],
            'Hy-mg': corr_mol_w['MgO'] + corr_mol_w['SiO2'],
            'Ol-mg': 2 * corr_mol_w['MgO'] + corr_mol_w['SiO2'],
            'Hm': corr_mol_w['Fe2O3'],
            'Ru': corr_mol_w['TiO2'],
            'Ap-F': 3 * corr_mol_w['CaO'] + corr_mol_w['P2O5'] + 1 / 3 * (
                    (corr_mol_w['CaO'] - _mol_w['O']) + 2 * corr_mol_w['F']),
            'Ap-O': 3 * corr_mol_w['CaO'] + corr_mol_w['P2O5'] + corr_mol_w['CaO'] * 1 / 3,
            'Cm': corr_mol_w['FeO'] + corr_mol_w['Cr2O3'],
            'Hl': (corr_mol_w['Na2O'] - _mol_w['O']) / 2 + _mol_w['Cl'],
            'Fr': (corr_mol_w['CaO'] - _mol_w['O']) + 2 * _mol_w['F'],
            'Pr': (corr_mol_w['FeO'] - _mol_w['O']) + 2 * _mol_w['S']
        }

        # 11 / Normative zircon
        if self.verbose > 1: print("Step 11 - Normative zircon")

        if minor_included:
            for i in data_n.index:
                if data_n.loc[i, 'SiO2'] > data_n.loc[i, 'ZrO2']:
                    n_phase.loc[i, 'Z'] = data_n.loc[i, 'ZrO2']
                    si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + n_phase.loc[i, 'Z']
                    data_n.loc[i, 'ZrO2'] = 0.
                else:
                    print('WARNING : No further SiO2 after zircon attribution for composition ' + str(
                        i) + '. Check data.')

        # 12 / Normative apatite
        if self.verbose > 1: print("Step 12 - Normative apatite")

        temp_ap = np.zeros(shape=[len(data_n.index), 1])
        for i in data_n.index:
            if data_n.loc[i, 'CaO'] >= (3 + 1 / 3) * data_n.loc[i, 'P2O5']:
                temp_ap[i] = data_n.loc[i, 'P2O5']
                data_n.loc[i, 'CaO'] = data_n.loc[i, 'CaO'] - (3 + 1 / 3) * data_n.loc[i, 'P2O5']
            else:
                temp_ap[i] = data_n.loc[i, 'CaO'] / (3 + 1 / 3)
                data_n.loc[i, 'CaO'] = 0.
                free.loc[i, 'P2O5'] = data_n.loc[i, 'P2O5'] - temp_ap[i]
            free.loc[i, 'P2O5'] = data_n.loc[i, 'P2O5'] - temp_ap[i]
            data_n.loc[i, 'P2O5'] = 0.

        n_phase['Ap-F'] = 0.
        n_phase['Ap-O'] = 0.
        if minor_included:
            for i in data_n.index:
                if data_n.loc[i, 'F'] >= 2 / 3 * temp_ap[i]:
                    n_phase.loc[i, 'Ap-F'] = temp_ap[i]
                    free.loc[i, 'O_12b'] = 1 / 3 * n_phase.loc[i, 'Ap-F']
                    data_n.loc[i, 'F'] = data_n.loc[i, 'F'] - 2 / 3 * n_phase.loc[i, 'Ap-F']
                else:
                    n_phase.loc[i, 'Ap-F'] = 1.5 * data_n.loc[i, 'F']
                    n_phase.loc[i, 'Ap-O'] = temp_ap[i] - 1.5 * data_n.loc[i, 'F']
                    free.loc[i, 'O_12c'] = data_n.loc[i, 'F'] / 2
                    data_n.loc[i, 'F'] = 0.
            n_phase['Ap'] = n_phase['Ap-F'] + n_phase['Ap-O']
        else:
            n_phase['Ap-O'] = temp_ap
        n_phase['Ap'] = n_phase['Ap-O'] + n_phase['Ap-F']

        # 13 / Normative fluorite
        if self.verbose > 1: print("Step 13 - Normative fluorite")

        if minor_included:
            for i in data_n.index:
                if data_n.loc[i, 'CaO'] >= data_n.loc[i, 'F'] / 2:
                    n_phase.loc[i, 'Fr'] = data_n.loc[i, 'F'] / 2
                    data_n.loc[i, 'CaO'] = data_n.loc[i, 'CaO'] - data_n.loc[i, 'F'] / 2
                    data_n.loc[i, 'F'] = 0.
                else:
                    n_phase.loc[i, 'Fr'] = data_n.loc[i, 'CaO']
                    data_n.loc[i, 'CaO'] = 0.
                    free.loc[i, 'F'] = data_n.loc[i, 'F'] - 2 * n_phase.loc[i, 'Fr']  # Unused F
                    data_n.loc[i, 'F'] = 0.
                free.loc[i, 'O_13'] = n_phase.loc[i, 'Fr']

        # 14 / Normative halite
        if self.verbose > 1: print("Step 14 - Normative halite")

        if minor_included and 'Cl' in data_n.keys():
            for i in data_n.index:
                if data_n.loc[i, 'Na2O'] >= 2 * data_n.loc[i, 'Cl']:
                    n_phase.loc[i, 'Hl'] = data_n.loc[i, 'Cl']
                    data_n.loc[i, 'Na2O'] = data_n.loc[i, 'Na2O'] - n_phase.loc[i, 'Hl'] / 2
                    free.loc[i, 'O_14'] = n_phase.loc[i, 'Hl'] / 2
                    data_n.loc[i, 'Cl'] = 0.
                else:
                    n_phase.loc[i, 'Hl'] = data_n.loc[i, 'Na2O'] / 2
                    free.loc[i, 'Cl'] = data_n.loc[i, 'Cl'] - n_phase.loc[i, 'Hl']
                    data_n.loc[i, 'Na2O'] = 0.

        # 15 / Normative thenardite
        if self.verbose > 1: print("Step 15 - Normative thenardite")

        if minor_included and 'SO3' in data_n.keys():
            for i in data_n.index:
                if data_n.loc[i, 'Na2O'] >= 2 * data_n.loc[i, 'SO3']:
                    n_phase.loc[i, 'Th'] = data_n.loc[i, 'SO3']
                    data_n.loc[i, 'Na2O'] = data_n.loc[i, 'Na2O'] - n_phase.loc[i, 'Th']
                    data_n.loc[i, 'SO3'] = 0.
                else:
                    n_phase.loc[i, 'Th'] = data_n.loc[i, 'Na2O']
                    free.loc[i, 'SO3'] = data_n.loc[i, 'SO3'] - n_phase.loc[i, 'Th']
                    data_n.loc[i, 'Na2O'] = 0.

        # 16 / Normative pyrite
        if self.verbose > 1: print("Step 16 - Normative pyrite")

        if minor_included and 'S' in data_n.keys():
            for i in data_n.index:
                if data_n.loc[i, 'FeO'] >= 2 * data_n.loc[i, 'S']:
                    n_phase.loc[i, 'Pr'] = data_n.loc[i, 'S'] / 2
                    data_n.loc[i, 'FeO'] = data_n.loc[i, 'FeO'] - n_phase.loc[i, 'Pr']
                    data_n.loc[i, 'S'] = 0.
                else:
                    n_phase.loc[i, 'Pr'] = data_n.loc[i, 'FeO']
                    free.loc[i, 'S'] = data_n.loc[i, 'S'] - 2 * n_phase.loc[i, 'Pr']
                    data_n.loc[i, 'FeO'] = 0.
                free.loc[i, 'O_16'] = n_phase.loc[i, 'Pr']

        # 17 / Normative sodium carbonate or calcite
        if self.verbose > 1: print("Step 17 - Normative sodium carbonate or calcite")

        if minor_included:
            if 'CO2' in data_keys:
                free['CO2'] = 0.
                if co2_cancrinite == co2_calcite == 0:
                    for i in data_n.index:
                        free.loc[i, 'CO2'] = data_n.loc[i, 'CO2']
                        data_n.loc[i, 'CO2'] = 0.
                else:
                    # Cancrinite
                    for i in data_n.index:
                        if data_n.loc[i, 'CO2'] > 0:
                            if data_n.loc[i, 'Na2O'] >= data_n.loc[i, 'CO2'] * co2_cancrinite:
                                n_phase.loc[i, 'Nc'] = data_n.loc[i, 'CO2'] * co2_cancrinite
                                data_n.loc[i, 'Na2O'] = data_n.loc[i, 'Na2O'] - n_phase.loc[i, 'Nc']
                            else:
                                n_phase.loc[i, 'Nc'] = data_n.loc[i, 'Na2O']
                                data_n.loc[i, 'Na2O'] = 0.
                                free.loc[i, 'CO2'] = free.loc[i, 'CO2'] + data_n.loc[i, 'CO2'] * co2_cancrinite \
                                                     - n_phase.loc[i, 'Nc']  # Free CO2
                            data_n.loc[i, 'CO2'] = data_n.loc[i, 'CO2'] * (1 - co2_cancrinite)

                    # Calcite, with all the rest of CO2
                    for i in data_n.index:
                        if data_n.loc[i, 'CO2'] > 0:
                            if data_n.loc[i, 'CaO'] > data_n.loc[i, 'CO2']:
                                n_phase.loc[i, 'Cc'] = data_n.loc[i, 'CO2']
                                data_n.loc[i, 'CaO'] = data_n.loc[i, 'CaO'] - n_phase.loc[i, 'Cc']
                            else:
                                n_phase.loc[i, 'Cc'] = data_n.loc[i, 'CaO']
                                data_n.loc[i, 'CaO'] = 0.
                                free.loc[i, 'CO2'] = free.loc[i, 'CO2'] + data_n.loc[i, 'CO2'] - n_phase.loc[
                                    i, 'Cc']  # CO2 to free
                            data_n.loc[i, 'CO2'] = 0.

        # 18 / Normative chromite
        if self.verbose > 1: print("Step 18 - Normative chromite")

        if minor_included:
            for i in data_n.index:
                if 0 < data_n.loc[i, 'Cr2O3'] <= data_n.loc[i, 'FeO']:
                    n_phase.loc[i, 'Cm'] = data_n.loc[i, 'Cr2O3']
                    data_n.loc[i, 'FeO'] = data_n.loc[i, 'FeO'] - n_phase.loc[i, 'Cm']
                    data_n.loc[i, 'Cr2O3'] = 0.
                elif data_n.loc[i, 'FeO'] < data_n.loc[i, 'Cr2O3']:
                    n_phase.loc[i, 'Cm'] = data_n.loc[i, 'FeO']
                    free.loc[i, 'Cr2O3'] = data_n.loc[i, 'Cr2O3'] - n_phase.loc[i, 'Cm']
                    data_n.loc[i, 'FeO'] = 0.
                    data_n.loc[i, 'Cr2O3'] = 0.

        # 19 / Normative ilmenite
        if self.verbose > 1: print("Step 19 - Normative ilmenite")

        for i in data_n.index:
            if data_n.loc[i, 'FeO'] >= data_n.loc[i, 'TiO2']:
                n_phase.loc[i, 'Il'] = data_n.loc[i, 'TiO2']
                data_n.loc[i, 'FeO'] = data_n.loc[i, 'FeO'] - n_phase.loc[i, 'Il']
                data_n.loc[i, 'TiO2'] = 0.
            else:
                n_phase.loc[i, 'Il'] = data_n.loc[i, 'FeO']
                data_n.loc[i, 'TiO2'] = data_n.loc[i, 'TiO2'] - n_phase.loc[i, 'Il']
                data_n.loc[i, 'FeO'] = 0.

        # 20 / Normative orthoclase or potassium metasilicate
        if self.verbose > 1: print("Step 20 - Normative orthoclase or potassium metasilicate")

        for i in data_n.index:
            if data_n.loc[i, 'Al2O3'] >= data_n.loc[i, 'K2O']:
                n_phase.loc[i, 'Orp'] = data_n.loc[i, 'K2O']
                data_n.loc[i, 'Al2O3'] = data_n.loc[i, 'Al2O3'] - n_phase.loc[i, 'Orp']
                data_n.loc[i, 'K2O'] = 0.
                si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 6 * n_phase.loc[i, 'Orp']
            else:
                n_phase.loc[i, 'Orp'] = data_n.loc[i, 'Al2O3']
                n_phase.loc[i, 'Ks'] = data_n.loc[i, 'K2O'] - n_phase.loc[i, 'Orp']  # Rest of K2O to Ks
                data_n.loc[i, 'Al2O3'] = 0.
                data_n.loc[i, 'K2O'] = 0.
                si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 6 * n_phase.loc[i, 'Orp'] + n_phase.loc[i, 'Ks']

        # 21 / Normative albite
        if self.verbose > 1: print("Step 21 - Normative albite")

        for i in data_n.index:
            if data_n.loc[i, 'Al2O3'] >= data_n.loc[i, 'Na2O']:
                n_phase.loc[i, 'Abp'] = data_n.loc[i, 'Na2O']
                data_n.loc[i, 'Al2O3'] = data_n.loc[i, 'Al2O3'] - n_phase.loc[i, 'Abp']
                data_n.loc[i, 'Na2O'] = 0.
            else:
                n_phase.loc[i, 'Abp'] = data_n.loc[i, 'Al2O3']
                data_n.loc[i, 'Na2O'] = data_n.loc[i, 'Na2O'] - n_phase.loc[i, 'Abp']  # Rest of Na2O to Ab
                data_n.loc[i, 'Al2O3'] = 0.
            si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 6 * n_phase.loc[i, 'Abp']

        # 22 / Normative acmite or sodium metasilicate
        if self.verbose > 1: print("Step 22 - Normative acmite or sodium metasilicate")

        for i in data_n.index:
            if data_n.loc[i, 'Na2O'] >= data_n.loc[i, 'Fe2O3']:
                n_phase.loc[i, 'Ac'] = data_n.loc[i, 'Fe2O3']
                n_phase.loc[i, 'Ns'] = data_n.loc[i, 'Na2O'] - n_phase.loc[i, 'Ac']  # Rest of Na2O to Ns
                data_n.loc[i, 'Na2O'] = 0.
                data_n.loc[i, 'Fe2O3'] = 0.
                si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 4 * n_phase.loc[i, 'Ac'] + n_phase.loc[i, 'Ns']
            else:
                n_phase.loc[i, 'Ac'] = data_n.loc[i, 'Na2O']
                data_n.loc[i, 'Fe2O3'] = data_n.loc[i, 'Fe2O3'] - n_phase.loc[i, 'Ac']
                data_n.loc[i, 'Na2O'] = 0.
                si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 4 * n_phase.loc[i, 'Ac']

        # 23 / Normative anorthite or corundum
        if self.verbose > 1: print("Step 23 - Normative anorthite or corundum")

        for i in data_n.index:
            if data_n.loc[i, 'Al2O3'] >= data_n.loc[i, 'CaO']:
                n_phase.loc[i, 'An'] = data_n.loc[i, 'CaO']
                n_phase.loc[i, 'C'] = data_n.loc[i, 'Al2O3'] - data_n.loc[i, 'CaO']  # Rest of Al2O3 in C
                data_n.loc[i, 'CaO'] = 0.
                data_n.loc[i, 'Al2O3'] = 0.
            else:
                n_phase.loc[i, 'An'] = data_n.loc[i, 'Al2O3']
                data_n.loc[i, 'CaO'] = data_n.loc[i, 'CaO'] - n_phase.loc[i, 'An']
                data_n.loc[i, 'Al2O3'] = 0.
            si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 2 * n_phase.loc[i, 'An']

        # 24 / Normative sphene / rutile
        if self.verbose > 1: print("Step 24 - Normative sphene / rutile")

        for i in data_n.index:
            if data_n.loc[i, 'CaO'] >= data_n.loc[i, 'TiO2']:
                n_phase.loc[i, 'Tnp'] = data_n.loc[i, 'TiO2']
                data_n.loc[i, 'CaO'] = data_n.loc[i, 'CaO'] - n_phase.loc[i, 'Tnp']
                data_n.loc[i, 'TiO2'] = 0.
            else:
                n_phase.loc[i, 'Tnp'] = data_n.loc[i, 'CaO']
                n_phase.loc[i, 'Ru'] = data_n.loc[i, 'TiO2'] - n_phase.loc[i, 'Tnp']  # Rest of TiO2 in Ru
                data_n.loc[i, 'CaO'] = 0.
                data_n.loc[i, 'TiO2'] = 0.
            si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + n_phase.loc[i, 'Tnp']

        # 25 / Normative magnetite or hematite
        if self.verbose > 1: print("Step 25 - Normative magnetite or hematite")

        for i in data_n.index:
            if data_n.loc[i, 'Fe2O3'] >= data_n.loc[i, 'FeO']:
                n_phase.loc[i, 'Mt'] = data_n.loc[i, 'FeO']
                n_phase.loc[i, 'Hm'] = data_n.loc[i, 'Fe2O3'] - n_phase.loc[i, 'Mt']  # Rest of Fe2O3 in Mt
                data_n.loc[i, 'FeO'] = 0.
                data_n.loc[i, 'Fe2O3'] = 0.
            else:
                n_phase.loc[i, 'Mt'] = data_n.loc[i, 'Fe2O3']
                data_n.loc[i, 'FeO'] = data_n.loc[i, 'FeO'] - n_phase.loc[i, 'Mt']
                data_n.loc[i, 'Fe2O3'] = 0.

        # 26 / Subdivision of Mg and Fe in some minerals
        if self.verbose > 1: print("Step 26 - Repartition of Mg and Fe in minerals")

        data_n['FeMgO'] = data_n['MgO'] + data_n['FeO']
        prop['xMg'] = data_n['MgO'] / (data_n['MgO'] + data_n['FeO'])
        prop['xFe'] = data_n['FeO'] / (data_n['MgO'] + data_n['FeO'])
        data_n['MgO'] = 0.
        data_n['FeO'] = 0.

        # 27 / Provisional normative diopside, wollastonite or hypersthene
        if self.verbose > 1: print("Step 27 - normative diopside, wollastonite or hypersthene")

        n_phase['Wop'] = 0.
        for i in data_n.index:
            if data_n.loc[i, 'CaO'] >= data_n.loc[i, 'FeMgO']:
                n_phase.loc[i, 'Dip'] = data_n.loc[i, 'FeMgO']
                n_phase.loc[i, 'Wop'] = data_n.loc[i, 'CaO'] - n_phase.loc[i, 'Dip']
                data_n.loc[i, 'CaO'] = 0.
                data_n.loc[i, 'FeMgO'] = 0.
                si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 2 * n_phase.loc[i, 'Dip'] + n_phase.loc[i, 'Wop']
                n_phase.loc[i, 'Hyp'] = 0.
            else:
                n_phase.loc[i, 'Dip'] = data_n.loc[i, 'CaO']
                n_phase.loc[i, 'Hyp'] = data_n.loc[i, 'FeMgO'] - n_phase.loc[i, 'Dip']
                data_n.loc[i, 'CaO'] = 0.
                data_n.loc[i, 'FeMgO'] = 0.
                si_deff.loc[i, 'Y'] = si_deff.loc[i, 'Y'] + 2 * n_phase.loc[i, 'Dip'] + n_phase.loc[i, 'Hyp']

        # 28 / Normative quartz and Si deficiency
        if self.verbose > 1: print("Step 28 - Normative quartz and Si deficiency")
        n_phase['Q'] = 0.
        for i in data_n.index:
            if data_n.loc[i, 'SiO2'] >= si_deff.loc[i, 'Y']:
                n_phase.loc[i, 'Q'] = data_n.loc[i, 'SiO2'] - si_deff.loc[i, 'Y']
                data_n.loc[i, 'SiO2'] = 0.
                next_step[i] = '36b'  # Instead of 36a which is obligatory in this algorithm
                if self.verbose: print("Si saturated for the composition", i)
            else:
                si_deff.loc[i, 'D'] = si_deff.loc[i, 'Y'] - data_n.loc[i, 'SiO2']
                data_n.loc[i, 'SiO2'] = 0.
                next_step[i] = '29'

        # 29 / Normative olivine or hypersthene
        if self.verbose > 1: print("Step 29 - Normative olivine or hypersthene")
        for i in data_n.index:
            if next_step[i] == '29':
                if si_deff.loc[i, 'D'] < n_phase.loc[i, 'Hyp'] / 2:
                    n_phase.loc[i, 'Ol'] = si_deff.loc[i, 'D']
                    n_phase.loc[i, 'Hy'] = n_phase.loc[i, 'Hyp'] - 2 * si_deff.loc[i, 'D']
                    next_step[i] = '36b'
                else:
                    n_phase.loc[i, 'Ol'] = n_phase.loc[i, 'Hyp'] / 2
                    n_phase.loc[i, 'Hy'] = 0.
                    si_deff.loc[i, 'D1'] = si_deff.loc[i, 'D'] - n_phase.loc[i, 'Hyp'] / 2
                    next_step[i] = '30'
            else:
                n_phase.loc[i, 'Hy'] = n_phase.loc[i, 'Hyp']
                n_phase.loc[i, 'Ol'] = 0.

        # 30 / Normative sphene or perovskite
        if self.verbose > 1: print("Step 30 - Normative sphene or perovskite")

        for i in data_n.index:
            if next_step[i] == '30':
                if si_deff.loc[i, 'D1'] < n_phase.loc[i, 'Tnp']:
                    n_phase.loc[i, 'Tn'] = n_phase.loc[i, 'Tnp'] - si_deff.loc[i, 'D1']
                    n_phase.loc[i, 'Pf'] = si_deff.loc[i, 'D1']
                    next_step[i] = '36c'
                else:
                    n_phase.loc[i, 'Pf'] = n_phase.loc[i, 'Tnp']
                    n_phase.loc[i, 'Tn'] = 0.
                    si_deff.loc[i, 'D2'] = si_deff.loc[i, 'D1'] - n_phase.loc[i, 'Tnp']
                    next_step[i] = '31'

        # 31 / Normative nepheline or albite
        if self.verbose > 1: print("Step 31 - Normative nepheline or albite")

        for i in data_n.index:
            if next_step[i] == '31':
                if si_deff.loc[i, 'D2'] < 4 * n_phase.loc[i, 'Abp']:
                    n_phase.loc[i, 'Ab'] = n_phase.loc[i, 'Abp'] - si_deff.loc[i, 'D2'] / 4
                    n_phase.loc[i, 'Ne'] = si_deff.loc[i, 'D2'] / 4
                    next_step[i] = '36d'
                else:
                    n_phase.loc[i, 'Ne'] = n_phase.loc[i, 'Abp']
                    n_phase.loc[i, 'Ab'] = 0
                    si_deff.loc[i, 'D3'] = si_deff.loc[i, 'D2'] - 4 * n_phase.loc[i, 'Abp']
                    next_step[i] = '32'

        # 32 / Normative leucite or orthoclase
        if self.verbose > 1: print("Step 32 - Normative leucite or orthoclase")

        n_phase['Lcp'] = 0.
        for i in data_n.index:
            if next_step[i] == '32':
                if si_deff.loc[i, 'D3'] < 2 * n_phase.loc[i, 'Orp']:
                    n_phase.loc[i, 'Or'] = n_phase.loc[i, 'Orp'] - si_deff.loc[i, 'D3'] / 2
                    n_phase.loc[i, 'Lc'] = si_deff.loc[i, 'D3'] / 2
                    next_step[i] = '36e'
                else:
                    n_phase.loc[i, 'Lcp'] = n_phase.loc[i, 'Orp']
                    n_phase.loc[i, 'Or'] = 0.
                    si_deff.loc[i, 'D4'] = si_deff.loc[i, 'D3'] - 2 * n_phase.loc[i, 'Orp']
                    next_step[i] = '33'

        # 33 / Normative dicalcium silicate or wollastonite
        if self.verbose > 1: print("Step 33 - Normative dicalcium silicate or wollastonite")

        for i in data_n.index:
            if next_step[i] == '33':
                if si_deff.loc[i, 'D4'] < 2 * n_phase.loc[i, 'Wop'] / 2:
                    n_phase.loc[i, 'Wo'] = n_phase.loc[i, 'Wop'] - 2 * si_deff.loc[i, 'D4']
                    n_phase.loc[i, 'Cs'] = si_deff.loc[i, 'D4']
                    next_step[i] = '36e'  # Instead of 36f
                else:
                    n_phase.loc[i, 'Cs'] = n_phase.loc[i, 'Wop'] / 2
                    n_phase.loc[i, 'Wo'] = 0.
                    si_deff.loc[i, 'D5'] = si_deff.loc[i, 'D4'] - n_phase.loc[i, 'Wop'] / 2
                    next_step[i] = '34'

        # 34 / Normative dicalcium silicate or olivine
        if self.verbose > 1: print("Step 34 - Normative diopside or olivine adjustment")

        for i in data_n.index:
            if next_step[i] == '34':
                if si_deff.loc[i, 'D5'] < n_phase.loc[i, 'Dip']:
                    n_phase.loc[i, 'Cs'] = n_phase.loc[i, 'Cs'] + si_deff.loc[i, 'D5'] / 2
                    n_phase.loc[i, 'Ol'] = n_phase.loc[i, 'Ol'] + si_deff.loc[i, 'D5'] / 2
                    n_phase.loc[i, 'Di'] = n_phase.loc[i, 'Dip'] - si_deff.loc[i, 'D5']
                    next_step[i] = '36g'
                else:
                    n_phase.loc[i, 'Cs'] = n_phase.loc[i, 'Cs'] + n_phase.loc[i, 'Dip'] / 2
                    n_phase.loc[i, 'Ol'] = n_phase.loc[i, 'Ol'] + n_phase.loc[i, 'Dip'] / 2
                    n_phase.loc[i, 'Di'] = n_phase.loc[i, 'Di-fe'] = n_phase.loc[i, 'Di-mg'] = 0.
                    si_deff.loc[i, 'D6'] = si_deff.loc[i, 'D5'] - n_phase.loc[i, 'Dip']
                    next_step[i] = '35'
            else:
                n_phase['Di'] = n_phase['Dip']

        # 35 / Normative kaliophilite or leucite
        if self.verbose > 1: print("Step 35 - Normative kaliophilite or leucite")

        for i in data_n.index:
            if next_step[i] == '35':
                if n_phase.loc[i, 'Lcp'] >= si_deff.loc[i, 'D6'] / 2:
                    n_phase.loc[i, 'Kp'] = si_deff.loc[i, 'D6'] / 2
                    n_phase.loc[i, 'Lc'] = n_phase.loc[i, 'Lcp'] - si_deff.loc[i, 'D6'] / 2
                    next_step[i] = '36g'
                else:
                    n_phase.loc[i, 'Kp'] = n_phase.loc[i, 'Lcp']
                    n_phase.loc[i, 'Lc'] = 0.
                    suppl.loc[i, 'defSiO2'] = (si_deff.loc[i, 'D6'] - 2 * n_phase.loc[i, 'Kp']) * corr_mol_w['SiO2']
                    next_step[i] = '36g'

        # 36 / Definite mineral proportions

        for i in data_n.index:
            if self.verbose > 1: print("Step 36 - Solution", i)

            if next_step[i] == '36b':
                if self.verbose > 1: print("     36b")
                n_phase.loc[i, 'Tn'] = n_phase.loc[i, 'Tnp']
                next_step[i] = '36c'

            if next_step[i] == '36c':
                if self.verbose > 1: print("     36c")
                n_phase.loc[i, 'Ab'] = n_phase.loc[i, 'Abp']
                next_step[i] = '36d'

            if next_step[i] == '36d':
                if self.verbose > 1: print("     36d")
                n_phase.loc[i, 'Or'] = n_phase.loc[i, 'Orp']
                n_phase.loc[i, 'Lc'] = n_phase.loc[i, 'Lcp']
                next_step[i] = '36e'

            if next_step[i] == '36e':
                if self.verbose > 1: print("     36e")
                if 'Wop' in n_phase.keys():
                    n_phase.loc[i, 'Wo'] = n_phase.loc[i, 'Wop']

        # Steps of 36a and 36f, obligatory for this algorithm
        if self.verbose > 1: print("Obligatory 36a, 36f and Fe-Mg distribution in olivine")
        n_phase['Hy-fe'] = n_phase['Hy'] * prop['xFe']
        n_phase['Hy-mg'] = n_phase['Hy'] * prop['xMg']
        n_phase['Di-fe'] = n_phase['Di'] * prop['xFe']
        n_phase['Di-mg'] = n_phase['Di'] * prop['xMg']
        n_phase['Ol-fe'] = n_phase['Ol'] * prop['xFe']
        n_phase['Ol-mg'] = n_phase['Ol'] * prop['xMg']

        # 37 / Conversion of normative minerals in %, normative sum
        if self.verbose > 1: print("Step 37 - Conversion of normative minerals in %, normative sum")

        ignored_phases = []
        for mineral in n_phase.keys():
            if mineral in final_list_min:
                partitions[mineral] = n_phase[mineral] * mol_w_min[mineral]
            elif mineral == 'Ap':
                suppl['Ap-F'] = n_phase['Ap-F'] * mol_w_min['Ap-F']
                suppl['Ap-O'] = n_phase['Ap-O'] * mol_w_min['Ap-O']
                partitions['Ap'] = suppl['Ap-F'] + suppl['Ap-O']
            elif mineral == 'Ol':
                suppl['Ol-mg'] = n_phase['Ol-mg'] * mol_w_min['Ol-mg']
                suppl['Ol-fe'] = n_phase['Ol-fe'] * mol_w_min['Ol-fe']
                partitions['Ol'] = suppl['Ol-mg'] + suppl['Ol-fe']
            elif mineral == 'Hy':
                suppl['Hy-mg'] = n_phase['Hy-mg'] * mol_w_min['Hy-mg']
                suppl['Hy-fe'] = n_phase['Hy-fe'] * mol_w_min['Hy-fe']
                partitions['Hy'] = suppl['Hy-mg'] + suppl['Hy-fe']
            elif mineral == 'Di':
                suppl['Di-mg'] = n_phase['Di-mg'] * mol_w_min['Di-mg']
                suppl['Di-fe'] = n_phase['Di-fe'] * mol_w_min['Di-fe']
                partitions['Di'] = suppl['Di-mg'] + suppl['Di-fe']
            else:
                ignored_phases.append(mineral)

        if self.verbose > 1: print("Temporary phases ignored in final composition :", *ignored_phases[skip_cols:])

        # Add free CO2
        free = free.fillna(0)
        free_phases = free.keys()
        if 'CO2' in free_phases:
            partitions['CO2'] = free['CO2'] * corr_mol_w['CO2']

        # Add free O
        if minor_included:
            # Free O
            free['O_wt%'] = 0.
            if 'O_12b' in free_phases:
                free['O_wt%'] = free['O_wt%'] + (1 + (0.1 * (mol_w_min['Ap-F'] / 328.8691887) - 1)) * \
                                corr_mol_w['O'] * free['O_12b']
            if 'O_12c' in free_phases:
                free['O_wt%'] = free['O_wt%'] + (
                        1 + 0.1 * (n_phase['Ap-F'] / n_phase['Ap']) * (mol_w_min['Ap-F'] / 328.8691887) - 1) * \
                                corr_mol_w['O'] * free['O_12c']
            if 'O_13' in free_phases:
                free['O_wt%'] = free['O_wt%'] + (1 + (corr_mol_w['CaO'] / 56.0774 - 1)) * corr_mol_w['O'] * free['O_13']
            if 'O_14' in free_phases:
                free['O_wt%'] = free['O_wt%'] + (1 + 0.5 * (corr_mol_w['Na2O'] / 61.97894 - 1)) * corr_mol_w['O'] * \
                                free['O_14']
            if 'O_16' in free_phases:
                free['O_wt%'] = free['O_wt%'] + (1 + (corr_mol_w['FeO'] / 71.8444 - 1)) * corr_mol_w['O'] * free['O_16']
            partitions['O'] = free['O_wt%']

        # Add free oxides
        partitions['free_Ox'] = 0.
        free_keys = free.keys()
        for oxel in ['P2O5', 'F', 'Cl', 'SO3', 'Cr2O3']:
            if oxel in free_keys:
                partitions['free_oxides'] = partitions['free_Ox'] + free[oxel] * _mol_w[oxel]

        # Totals
        partitions['Sum_norm'] = partitions.iloc[:, skip_cols:].sum(axis=1)
        partitions = partitions.round(to_round)

        # 38 / Correctness of normative sum
        if self.verbose > 1: print("Step 38 - Correctness of normative sum")

        suppl['diff_sum'] = total_normalized - partitions['Sum_norm']

        # 39 / Other petrogenetical parameters
        if self.verbose > 1: print("Step 39 - Other petrogenetical parameters")

        suppl['pp_salic'] = n_phase['Q'] + n_phase['Or'] + n_phase['Ab'] + n_phase['An']
        suppl['pp_femic'] = suppl['Di-mg'] + suppl['Di-fe'] + suppl['Hy-mg'] + suppl['Hy-fe'] + suppl['Ol-mg'] + \
                            n_phase['Ol-fe'] + n_phase['Mt'] + n_phase['Il'] + n_phase['Hm']
        suppl['pp_CI'] = n_phase['An'] + 2.1570577 * suppl['Di-mg'] + suppl['Ol-mg'] + 0.7007616 * suppl['Hy-fe']
        suppl['pp_DI'] = n_phase['Q'] + n_phase['Or'] + n_phase['Ab'] + n_phase['An']

        partitions = partitions.fillna(0)
        partitions = partitions.loc[:, (partitions != 0).any(axis=0)]
        free = free.fillna(0)
        free = free.round(to_round)
        free = free.loc[:, (free != 0).any(axis=0)]
        suppl = suppl.round(to_round)
        suppl = suppl.fillna(0)

        if self.verbose:
            print(">>> Final compositions (wt %)")
            print(partitions.to_string())
            print(">>> Free phases (mol)")
            print(free.to_string())
            print(">>> Supplementary data")
            print(suppl.to_string())

        return partitions, (free, suppl)
